__all__ = ('ConverterFlag', )

import re
from datetime import timedelta

from ...env import CACHE_USER
from ...backend.utils import cached_property, copy_docs, FunctionType
from ...backend.analyzer import CallableAnalyzer
from ...discord.utils import USER_MENTION_RP, CHANNEL_MENTION_RP, ID_RP, parse_tdelta, parse_rdelta, \
    INVITE_CODE_RP, CHANNEL_MESSAGE_RP
from ...discord.bases import FlagBase
from ...discord.guild import Guild
from ...discord.core import USERS, CLIENTS, ROLES, CHANNELS, EMOJIS, GUILDS, MESSAGES
from ...discord.exceptions import DiscordException, ERROR_CODES
from ...discord.user import User, UserBase
from ...discord.channel import ChannelGuildBase, ChannelBase, ChannelTextBase, ChannelText, ChannelPrivate, \
    ChannelVoice, ChannelGroup, ChannelCategory, ChannelStore, ChannelThread, ChannelVoiceBase, ChannelStage, \
    ChannelDirectory
from ...discord.client import Client
from ...discord.emoji import Emoji, parse_emoji
from ...discord.invite import Invite
from ...discord.role import Role, parse_role_mention
from ...discord.color import Color, parse_color
from ...discord.http import MESSAGE_JUMP_URL_RP, INVITE_URL_RP
from ...discord.message import Message
from ...discord.sticker import Sticker

from .exceptions import CommandParameterParsingError
from .context import CommandContext
from .utils import raw_name_to_display

try:
    from dateutil.relativedelta import relativedelta
except ImportError:
    relativedelta = None


NUMERIC_CONVERSION_LIMIT = 100

CONTENT_ARGUMENT_PARSERS = {}

DEFAULT_PARAMETER_SEPARATOR = ('"', '"')
DEFAULT_PARAMETER_ASSIGNER = ':'


class ContentParameterParserContextBase:
    """
    Parsing context returned by ``ContentParameterParser``.
    
    Attributes
    ----------
    _cache : `dict` of (`str`, `Any`)
        Cache used by cached properties.
    _parsed : `re.Match`
        The parsed regex.
    """
    __slots__ = ('_cache', '_parsed')
    def __new__(cls, parsed):
        """
        Creates a new ``ContentParameterParserContext`` instance with the given match.
        
        Parameters
        ----------
        parsed : `re.Match`
            The parsed regex.
        """
        self = object.__new__(cls)
        self._parsed = parsed
        self._cache = {}
        return self
    
    @property
    def has_keyword(self):
        """
        Returns whether keyword could be parsed.
        
        Returns
        -------
        has_keyword : `bool`
        """
        return (self.keyword is not None)
    
    @property
    def end(self):
        """
        Returns the end of the parsed region.
        
        Returns
        -------
        end : `str`
        """
        return self._parsed.end()
    
    @cached_property
    def whole(self):
        """
        Gets the whole parsed string part.
        
        Returns
        -------
        whole : `str`
        """
        return ''
    
    @cached_property
    def keyword(self):
        """
        Gets the parsed keyword from the parsed part.
        
        Returns
        -------
        keyword : `None` or `str`
        """
        return None
    
    @cached_property
    def value(self):
        """
        Returns the parsed value.
        
        Returns
        -------
        value : `str`
        """
        return ''

class ContentParameterParserContextSeparator(ContentParameterParserContextBase):
    """
    Separator pattern based parsing context returned by ``ContentParameterParser``.
    
    Attributes
    ----------
    _cache : `dict` of (`str`, `Any`)
        Cache used by cached properties.
    _parsed : `re.Match`
        The parsed regex.
    """
    __slots__ = ()
    
    @cached_property
    @copy_docs(ContentParameterParserContextBase.whole)
    def whole(self):
        return self._parsed.group(1)
    
    @cached_property
    @copy_docs(ContentParameterParserContextBase.keyword)
    def keyword(self):
        return self._parsed.group(2)
    
    @cached_property
    @copy_docs(ContentParameterParserContextBase.value)
    def value(self):
        return self._parsed.group(3)

class ContentParameterParserContextEncapsulator(ContentParameterParserContextBase):
    """
    Encapsulator pattern based parsing context returned by ``ContentParameterParser``.
    
    Attributes
    ----------
    _cache : `dict` of (`str`, `Any`)
        Cache used by cached properties.
    _parsed : `re.Match`
        The parsed regex.
    """
    __slots__ = ()
    
    @cached_property
    @copy_docs(ContentParameterParserContextBase.whole)
    def whole(self):
        parsed = self._parsed
        part = parsed.group(3)
        if part is None:
            part = parsed.group(2)
        
        return part
    
    @cached_property
    @copy_docs(ContentParameterParserContextBase.keyword)
    def keyword(self):
        return self._parsed.group(1)
    
    @cached_property
    @copy_docs(ContentParameterParserContextBase.value)
    def value(self):
        parsed = self._parsed
        part = parsed.group(3)
        if part is None:
            part = parsed.group(2)
        
        return part


class ContentParameterParser:
    """
    Content parameter parser used inside of a ``ContentParserContext`` and stored by ``CommandContentParser``
    instances.
    
    Attributes
    ----------
    _context_class : ``ContentParameterParserContextBase``
        Context class to interact with the parsed string.
    _rp : `_sre.SRE_Pattern`
        The regex pattern what is passed and used by the caller.
    assigner : `str`
        Assigner executed by the ``ContentParameterParser`` instance.
    separator : `str` or `tuple` (`str`, `str`)
        The executed separator by the ``ContentParameterParser`` instance.
    """
    __slots__ = ('_context_class', '_rp', 'assigner', 'separator')
    def __new__(cls, separator, assigner):
        """
        Creates a new ``ContentParameterSeparator`` instance. If one already exists with the given parameters, returns
        that instead.
        
        Parameters
        ----------
        separator : `None`, `str`, `tuple` (`str`, `str`)
            The executed separator by the ``ContentParameterSeparator`` instance.
        assigner : `None`, `str`
            The assigner for keyword-only parameters.
        
        Raises
        ------
        TypeError
            - If `separator` is not given as `None`, `str`, neither as `tuple` instance.
            - If `separator` was given as `tuple`, but it's element are not `str` instances.
            - If `assigner` was not given neither as `None` or as `str` instance.
        ValueError
            - If `separator` is given as `str`, but it's length is not `1`.
            - If `separator` is given as `str`, but it is a space character.
            - If `separator` is given as `tuple`, but one of it's element's length is not `1`.
            - If `separator` is given as `tuple`, but one of it's element's is a space character.
            - If `assigner`'s length is not `1`.
        """
        if separator is None:
            separator = DEFAULT_PARAMETER_SEPARATOR
            separator_type = type(separator)
        else:
            separator_type = type(separator)
            if separator_type is str:
                processed_separator = separator
            elif separator_type is tuple:
                processed_separator = list(separator)
            elif issubclass(separator_type, str):
                processed_separator = str(separator)
                separator_type = str
            elif issubclass(separator_type, tuple):
                processed_separator = list(separator)
                separator_type = tuple
            else:
                raise TypeError(f'`separator` can be given as `None`, `str` or as `tuple` instance, got '
                    f'{separator_type.__name__}.')
            
            if separator_type is str:
                if len(processed_separator) != 1:
                    raise ValueError(f'`str` separator length can be only `1`, got {separator!r}.')
                
                if processed_separator.isspace():
                    raise ValueError(f'`str` separator cannot be a space character`, meanwhile it is, got '
                        f'{separator!r}.')
                
                separator = processed_separator
            
            else:
                if len(processed_separator) != 2:
                    raise ValueError(f'`tuple` separator length can be only `2`, got {separator!r}.')
                
                for index in range(2):
                    element = processed_separator[index]
                    
                    element_type = element.__class__
                    if element_type is str:
                        processed_element = element
                    elif issubclass(element_type, str):
                        processed_element = str(element)
                        processed_separator[index] = processed_element
                    else:
                        raise TypeError(f'`tuple` separator\'s elements can be only `str` instances, meanwhile it\'s '
                            f'element under index `{index}` is type {element_type.__name__!r}.')
                    
                    if len(processed_element) != 1:
                        raise ValueError(f'`tuple` separator\'s elements can be only `str` with length of `1`, '
                            f'meanwhile it\'s element under index `{index}` is not, got {element!r}.')
                    
                    if processed_element.isspace():
                        raise ValueError(f'`tuple` separator\'s elements cannot be space character`, meanwhile it\'s '
                            f'element under index `{index}` is, got {element!r}.')
                
                separator = tuple(processed_separator)
        
        if assigner is None:
            assigner = DEFAULT_PARAMETER_ASSIGNER
        else:
            assigner_type = type(assigner)
            if assigner_type is str:
                pass
            elif issubclass(assigner_type, str):
                assigner = str(assigner)
            else:
                raise TypeError(f'`assigner` can be given as `None` or `str` instance, got {assigner_type.__name__}.')
        
        if len(assigner) != 1:
            raise ValueError(f'`assigner` length can be `1`, got {len(assigner)}; {assigner!r}.')
        
        try:
            return CONTENT_ARGUMENT_PARSERS[(separator, assigner)]
        except KeyError:
            pass
        
        assigner_escaped = re.escape(assigner)
        if separator_type is str:
            escaped_separator = re.escape(separator)
            rp = re.compile(f'[{escaped_separator}\s]*((?:([^\s{assigner_escaped}]+?)\s*{assigner_escaped}\s+)?(.+?))\s*(?:$|[{escaped_separator})]+)', re.M|re.S)
            
            context_class = ContentParameterParserContextSeparator
        else:
            start, end = separator
            if start == end:
                escaped_separator = re.escape(start)
                rp = re.compile(f'\s*(?:([^\s{assigner_escaped}]+?)\s*{assigner_escaped}\s+)?(?:(?:{escaped_separator}(.+?)(?:$|{escaped_separator}))|(?:(.+?)(?:$|[{escaped_separator}\s]+)))', re.M|re.S)
            
            else:
                separator_start_escaped = re.escape(start)
                separator_end_escaped = re.escape(end)
                rp = re.compile(f'\s*(?:([^\s{assigner_escaped}]+?)\s*{assigner_escaped}\s+)?(?:(?:{separator_start_escaped}(.+?)(?:$|{separator_end_escaped}))|(?:(.+?)(?:$|[{separator_start_escaped}\s]+)))', re.M|re.S)
            
            context_class = ContentParameterParserContextEncapsulator
        
        self = object.__new__(cls)
        self.assigner = assigner
        self.separator = separator
        self._rp = rp
        self._context_class = context_class
        
        CONTENT_ARGUMENT_PARSERS[(separator, assigner)] = self
        return self
    
    def __call__(self, content, index):
        """
        Calls the content parameter separator to get the next part of the given content.
        
        Parameters
        ----------
        content : `str`
            The content what's next part we are going to be parsed.
        index : `int`
            The starter index of the content to parse from.
        
        Returns
        -------
        keyword : `None` or `str`
            The parsed out part's keyword if any.
        value : `str`
            The parsed out value.
        end : `int`
            The index where the next parsing should start from.
        """
        context = self._context_class(self._rp.match(content, index))
        return context.keyword, context.value, context.end
    
    def __repr__(self):
        """Returns the content parameter separator's representation."""
        return f'{self.__class__.__name__}({self.separator!r}, {self.assigner!r})'
    
    def __hash__(self):
        """Returns the content parameter parser's hash."""
        return hash(self.separator) ^ hash(self.assigner)
    
    def __eq__(self, other):
        """Returns whether the two content parameter separator are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.separator != other.separator:
            return False
        
        if self.assigner != other.assigner:
            return False
        
        return True


DEFAULT_SEPARATOR = ContentParameterParser(DEFAULT_PARAMETER_SEPARATOR, DEFAULT_PARAMETER_ASSIGNER)


def parse_user_mention(part, message):
    """
    If the message's given part is a user mention, returns the respective user.
    
    Parameters
    ----------
    part : `str`
        A part of a message's content.
    message : ``Message``
        The respective message of the given content part.
    
    Returns
    -------
    user : `None` or ``UserBase`` instance
    """
    user_mentions = message.user_mentions
    if user_mentions is None:
        return
    
    parsed = USER_MENTION_RP.fullmatch(part)
    if parsed is None:
        return
    
    user_id = int(parsed.group(1))
    return USERS.get(user_id, None)


def parse_channel_mention(part, message):
    """
    If the message's given part is a channel mention, returns the respective channel.
    
    Parameters
    ----------
    part : `str`
        A part of a message's content.
    message : ``Message``
        The respective message of the given content part.
        
    Returns
    -------
    channel : `None` or ``ChannelBase`` instance
    """
    channel_mentions = message.channel_mentions
    if channel_mentions is None:
        return
    
    parsed = CHANNEL_MENTION_RP.fullmatch(part)
    if parsed is None:
        return

    channel_id = int(parsed.group(1))
    for channel in channel_mentions:
        if channel.id == channel_id:
            return channel

REST_PARSER_RP = re.compile('\s*(.*?)\s*', re.M|re.S)


def parse_rest_content(content, index):
    """
    Parses everything till the content's end.
    
    Parameters
    ----------
    content : `str`
        A message's content after a respective prefix.
    index : `int`
        The start till we should cut down the content.
    
    Returns
    -------
    rest : `str`
        The content's end.
    """
    return REST_PARSER_RP.fullmatch(content, index).group(1)


class ConverterFlag(FlagBase):
    """
    Flags for a converter to describe by which rules it should convert.
    
    +---------------+---------------+-----------------------------------------------------------------------+
    | Rule name     | Shift value   | Description                                                           |
    +===============+===============+=======================================================================+
    | url           | 0             | Whether the entity should be parsed from it's url.                    |
    +---------------+---------------+-----------------------------------------------------------------------+
    | mention       | 1             | Whether the entity should be parsed out from it's mention.            |
    +---------------+---------------+-----------------------------------------------------------------------+
    | name          | 2             | Whether the entity should be picked up by it's name.                  |
    +---------------+---------------+-----------------------------------------------------------------------+
    | id            | 3             | Whether the entity should be picked up by it's name.                  |
    +---------------+---------------+-----------------------------------------------------------------------+
    | everywhere    | 4             | Whether the entity should be searched out of the local scope.         |
    |               |               | Mostly pairs with the `id` flag.                                      |
    +---------------+---------------+-----------------------------------------------------------------------+
    | profile       | 5             | User parser only. Can be used when user cache is disabled to          |
    |               |               | ensure, that the user will have local guild profile if applicable.    |
    +---------------+---------------+-----------------------------------------------------------------------+
    
    There are already precreated flags, which are the following:
    +-------------------+-------------------------------------------+
    | Name              | Included flags                            |
    +===================+===========================================+
    | user_default      | mention, name, id                         |
    +-------------------+-------------------------------------------+
    | user_all          | mention, name, id, everywhere, profile    |
    +-------------------+-------------------------------------------+
    | client_default    | mention, name, id                         |
    +-------------------+-------------------------------------------+
    | client_all        | mention, name, id, everywhere             |
    +-------------------+-------------------------------------------+
    | role_default      | mention, name, id                         |
    +-------------------+-------------------------------------------+
    | role_all          | mention, name, id, everywhere             |
    +-------------------+-------------------------------------------+
    | channel_default   | mention, name, id                         |
    +-------------------+-------------------------------------------+
    | channel_all       | mention, name, id, everywhere             |
    +-------------------+-------------------------------------------+
    | emoji_default     | mention, name, id                         |
    +-------------------+-------------------------------------------+
    | emoji_all         | mention, name, id, everywhere             |
    +-------------------+-------------------------------------------+
    | guild_default     | id                                        |
    +-------------------+-------------------------------------------+
    | guild_all         | id, everywhere                            |
    +-------------------+-------------------------------------------+
    | message_default   | url, id                                   |
    +-------------------+-------------------------------------------+
    | message_all       | url, id, everywhere                       |
    +-------------------+-------------------------------------------+
    | invite_default    | url, id                                   |
    +-------------------+-------------------------------------------+
    | invite_all        | url, id                                   |
    +-------------------+-------------------------------------------+
    | sticker_default   | name, id                                  |
    +-------------------+-------------------------------------------+
    | sticker_all       | name, id, everywhere                      |
    +-------------------+-------------------------------------------+
    
    Note, if you use for example a `'user'` parser, then by default it will use the `user_default` flags, and it
    will ignore everything else, than `user_all`.
    
    Some events, like `int`, or `str` do not have any flags, what means, their behaviour cannot be altered.
    """
    __keys__ = {
        'url': 0,
        'mention': 1,
        'name': 2,
        'id': 3,
        'everywhere': 4,
        'profile': 5,
    }
    
    user_default = NotImplemented
    user_all = NotImplemented
    client_default = NotImplemented
    client_all = NotImplemented
    role_default = NotImplemented
    role_all = NotImplemented
    channel_default = NotImplemented
    channel_all = NotImplemented
    emoji_default = NotImplemented
    emoji_all = NotImplemented
    guild_default = NotImplemented
    guild_all = NotImplemented
    message_default = NotImplemented
    message_all = NotImplemented
    invite_default = NotImplemented
    invite_all = NotImplemented
    sticker_default = NotImplemented
    sticker_all = NotImplemented

ConverterFlag.user_default = ConverterFlag().update_by_keys(mention=True, name=True, id=True)
ConverterFlag.user_all = ConverterFlag.user_default.update_by_keys(everywhere=True, profile=True)
ConverterFlag.client_default = ConverterFlag().update_by_keys(mention=True, name=True, id=True)
ConverterFlag.client_all = ConverterFlag.client_default.update_by_keys(everywhere=True)
ConverterFlag.role_default = ConverterFlag().update_by_keys(mention=True, name=True, id=True)
ConverterFlag.role_all = ConverterFlag.role_default.update_by_keys(everywhere=True)
ConverterFlag.channel_default = ConverterFlag().update_by_keys(mention=True, name=True, id=True)
ConverterFlag.channel_all = ConverterFlag.channel_default.update_by_keys(everywhere=True)
ConverterFlag.emoji_default = ConverterFlag().update_by_keys(mention=True, name=True, id=True)
ConverterFlag.emoji_all = ConverterFlag.emoji_default.update_by_keys(everywhere=True)
ConverterFlag.guild_default = ConverterFlag().update_by_keys(id=True)
ConverterFlag.guild_all = ConverterFlag.guild_default.update_by_keys(everywhere=True)
ConverterFlag.message_default = ConverterFlag().update_by_keys(url=True, id=True)
ConverterFlag.message_all = ConverterFlag.message_default.update_by_keys(everywhere=True)
ConverterFlag.invite_default = ConverterFlag().update_by_keys(url=True, id=True)
ConverterFlag.invite_all = ConverterFlag.invite_default
ConverterFlag.sticker_default = ConverterFlag().update_by_keys(name=True, id=True)
ConverterFlag.sticker_all = ConverterFlag.sticker_default.update_by_keys(everywhere=True)


CONVERTER_FLAG_URL = 1 << ConverterFlag.__keys__['url']
CONVERTER_FLAG_MENTION = 1 << ConverterFlag.__keys__['mention']
CONVERTER_FLAG_NAME = 1 << ConverterFlag.__keys__['name']
CONVERTER_FLAG_ID = 1 << ConverterFlag.__keys__['id']
CONVERTER_FLAG_EVERYWHERE = 1 << ConverterFlag.__keys__['everywhere']
CONVERTER_FLAG_PROFILE = 1 << ConverterFlag.__keys__['profile']



CONVERTER_SETTING_TYPE_TO_SETTING = {}
CONVERTER_SETTING_NAME_TO_SETTING = {}
CONVERTER_NAME_TO_TYPE = {}


class ConverterSetting:
    """
    Store settings about a converter.
    
    Attributes
    ----------
    all_flags : ``ConverterFlag``
        All the flags which the converter picks up.
    alternative_type_name : `None` or `str`
        Alternative string name for the parser, which allows picking up a respective converter.
    alternative_types : `None` or `list` of `type` instances
        Alternative type specifications, which are supported by the parser.
    converter : `async-function`
        The converter function.
    default_flags : ``ConverterFlag``
        The default flags with what the converter will be used if not defining any specific.
    default_type : `None` or `type`
        The default annotation type of the converter.
    requires_part : `bool`
        Whether the converter requires data to parse or it just produces it's result on the fly.
    uses_flags : `bool`
        Whether the converter processes any flags.
    """
    __slots__ = ('all_flags', 'alternative_type_name', 'alternative_types', 'converter', 'default_flags',
        'default_type', 'requires_part', 'uses_flags')
    
    def __new__(cls, converter, uses_flags, default_flags, all_flags, alternative_type_name, default_type,
            alternative_types, requires_part):
        """
        Creates a new ``ConverterSetting`` instance to store settings related to a converter function.
        
        Parameters
        ----------
        converter : `function` (async)
            The converter function.
        uses_flags : `bool`
            Whether the converter processes any flags.
        default_flags : ``ConverterFlag``
            The default flags with what the converter will be used if not defining any specific.
        all_flags : ``ConverterFlag``
             All the flags which the converter picks up.
        alternative_type_name : `None` or `str`
            Alternative string name for the parser, what allows picking up a respective converter.
        default_type : `None` or `type`
            The default annotation type of the converter.
        alternative_types : `None` `iterable` of `type`
            A list of the alternatively accepted types.
        requires_part : `bool`
            Whether the converter requires data to parse or it just produces it's result on the fly.
        
        Raises
        -------
        TypeError
            - If `converter` is not `function` type.
            - If `converter` is not `async`.
            - If `converter` accepts bad amount of parameters.
            - If `uses_flags` was not given as `bool`, nether as `int` as `0` or `1`.
            - If `default_flags` or `all_flags` was not given as `ConverterFlag` instance.
            - If `alternative_type_name` was not given as `None`, neither as `str` instance.
            - If `default_type` was not given as `None`, neither as `type` instance.
            - If `alternative_types` was not given as `None`, neither as `iterable` of `type`.
            - If `requires_part` was not to given as `bool` instance.
        ValueError
            If `uses_flags` is given as `true`, but at the same time `all_flags` was not given as
            `ConverterFlag(0)`
        """
        if not isinstance(converter, FunctionType):
            raise TypeError(f'`converter` should have been given as `{FunctionType.__name__}` instance, got '
                f'{converter.__class__.__name__}.')
        
        if type(requires_part) is not bool:
            raise TypeError(f'`requires_part` can be given as `bool`, got {requires_part.__class__.__name__}.')
        
        analyzed = CallableAnalyzer(converter)
        if (not analyzed.is_async()):
            raise TypeError(f'`converter` should have been given as an async function instance, got '
                f'{converter!r}.')
        
        non_reserved_positional_parameter_count = analyzed.get_non_reserved_positional_parameter_count()
        if non_reserved_positional_parameter_count != (3 if requires_part else 2):
            raise TypeError(f'`converter` should accept `3` (or 2 if `requires_part` is `False`) non reserved '
                f'positional parameters , meanwhile it expects {non_reserved_positional_parameter_count}.')
        
        if analyzed.accepts_args():
            raise TypeError(f'`converter` should accept not expect args, meanwhile it does.')
        
        if analyzed.accepts_kwargs():
            raise TypeError(f'`converter` should accept not expect kwargs, meanwhile it does.')
        
        non_default_keyword_only_parameter_count = analyzed.get_non_default_keyword_only_parameter_count()
        if non_default_keyword_only_parameter_count:
            raise TypeError(f'`converter` should accept `0` keyword only parameters, meanwhile it expects '
                f'{non_default_keyword_only_parameter_count}.')
        
        if type(uses_flags) is not bool:
            raise TypeError(f'`uses_flags` can be given as `bool`, got {uses_flags.__class__.__name__}.')
        
        if type(default_flags) is not ConverterFlag:
            raise TypeError(f'`default_flags` should have be given as `{ConverterFlag.__name__}`, got '
                f'{default_flags.__class__.__name__}.')
        
        if type(all_flags) is not ConverterFlag:
            raise TypeError(f'`all_flags` should have be given as `{ConverterFlag.__name__}`, got '
                f'{all_flags.__class__.__name__}.')
        
        if (alternative_type_name is not None) and (type(alternative_type_name) is not str):
            raise TypeError(f'`alternative_type_name` should have be given as `None` or as `str`, got '
                f'{alternative_type_name.__class__.__name__}.')
        
        if (default_type is not None) and (not isinstance(default_type, type)):
            raise TypeError(f'`default_type` was not given as `None`, neither as `type` instance, got '
                f'{default_type.__class__.__name__}.')
        
        if (alternative_types is None):
            alternative_types_processed = None
        
        else:
            alternative_types_type = alternative_types.__class__
            if not hasattr(alternative_types_type, '__iter__'):
                raise TypeError(f'`alternative_types` can be given as `None` or as `iterable` of `type`, got '
                    f'{alternative_types_type.__name__}.')
            
            alternative_types_processed = []
            
            index = 1
            for alternative_type in alternative_types:
                if not isinstance(alternative_type, type):
                    raise TypeError(f'`alternative_types` can be given as `None`, or as `iterable` of `type`, got '
                        f'`iterable`, but it\'s {index} element is {alternative_type.__class__.__name__}.')
                
                alternative_types_processed.append(alternative_type)
                index +=1
            
            if not alternative_types_processed:
                 alternative_types_processed = None
        
        if (not uses_flags) and all_flags:
            raise ValueError(f'If `uses_flags` is given as `true`, then `all_flags` should be given as '
                f'`{ConverterFlag.__name__}(0)`, got {all_flags!r}.')
        
        self = object.__new__(cls)
        self.converter = converter
        self.uses_flags = uses_flags
        self.default_flags = default_flags
        self.all_flags = all_flags
        self.alternative_type_name = alternative_type_name
        self.default_type = default_type
        self.alternative_types = alternative_types_processed
        self.requires_part = requires_part
        
        if (default_type is not None):
            CONVERTER_SETTING_TYPE_TO_SETTING[default_type] = self
            CONVERTER_SETTING_NAME_TO_SETTING[default_type.__name__] = self
            CONVERTER_NAME_TO_TYPE[default_type.__name__] = default_type
        
        if (alternative_type_name is not None):
            CONVERTER_SETTING_NAME_TO_SETTING[alternative_type_name] = self
            CONVERTER_NAME_TO_TYPE[alternative_type_name] = default_type
        
        if (alternative_types_processed is not None):
            for alternative_type in alternative_types_processed:
                CONVERTER_SETTING_TYPE_TO_SETTING[alternative_type] = self
                CONVERTER_SETTING_NAME_TO_SETTING[alternative_type.__name__] = self
                CONVERTER_NAME_TO_TYPE[alternative_type.__name__] = alternative_type
        
        return self
    
    def __repr__(self):
        """Returns the converter setting's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' converter=',
            self.converter.__name__,
        ]
        
        default_type = self.default_type
        if default_type is None:
            alternative_type_name = self.alternative_type_name
            if (alternative_type_name is not None):
                repr_parts.append(', alternative_type_name=')
                repr_parts.append(repr(alternative_type_name))
        else:
            default_type_name = default_type.__name__
            repr_parts.append(', default_type=')
            repr_parts.append(default_type_name)
            
            alternative_type_name = self.alternative_type_name
            if alternative_type_name != default_type_name:
                repr_parts.append(', alternative_type_name=')
                repr_parts.append(repr(alternative_type_name))
            
            alternative_types = self.alternative_types
            if (alternative_types is not None):
                repr_parts.append(', alternative_types=[')
                
                index = 0
                limit = len(alternative_types)
                while True:
                    alternative_type_= alternative_types[index]
                    index += 1
                    
                    repr_parts.append(alternative_type_.__name__)
                    
                    if index == limit:
                        break
                    
                    repr_parts.append(', ')
                    continue
                
                repr_parts.append(']')
        
        if self.uses_flags:
            default_flags = self.default_flags
            repr_parts.append(', default_flags=')
            repr_parts.append(int.__repr__(default_flags))
            
            all_flags = self.all_flags
            if default_flags != all_flags:
                repr_parts.append(', all_flags=')
                repr_parts.append(int.__repr__(all_flags))
        
        repr_parts.append('>')
        return ''.join(repr_parts)


async def none_converter(command_context, content_parser_parameter_detail):
    return None


CONVERTER_NONE = ConverterSetting(
    converter = none_converter,
    uses_flags = False,
    default_flags = ConverterFlag(),
    all_flags = ConverterFlag(),
    alternative_type_name = None,
    default_type = None,
    alternative_types = None,
    requires_part = False
)


async def command_context_converter(command_context, content_parser_parameter_detail):
    return command_context

CONVERTER_SELF_CONTEXT = ConverterSetting(
    converter = command_context_converter,
    uses_flags = False,
    default_flags = ConverterFlag(),
    all_flags = ConverterFlag(),
    alternative_type_name = None,
    default_type = CommandContext,
    alternative_types = None,
    requires_part = False
)


async def self_client_converter(command_context, content_parser_parameter_detail):
    return command_context.client

CONVERTER_SELF_CLIENT = ConverterSetting(
    converter = self_client_converter,
    uses_flags = False,
    default_flags = ConverterFlag(),
    all_flags = ConverterFlag(),
    alternative_type_name = None,
    default_type = None,
    alternative_types = None,
    requires_part = False
)


async def self_message_converter(command_context, content_parser_parameter_detail):
    return command_context.message

CONVERTER_SELF_MESSAGE = ConverterSetting(
    converter = self_message_converter,
    uses_flags = False,
    default_flags = ConverterFlag(),
    all_flags = ConverterFlag(),
    alternative_type_name = None,
    default_type = None,
    alternative_types = None,
    requires_part = False
)



if CACHE_USER:
    async def user_converter(command_context, content_parser_parameter_detail, part):
        flags = content_parser_parameter_detail.flags
        message = command_context.message
        
        if flags&CONVERTER_FLAG_ID:
            parsed = ID_RP.fullmatch(part)
            if (parsed is not None):
                id_ = int(parsed.group(1))
                
                if flags&CONVERTER_FLAG_EVERYWHERE:
                    try:
                        user = USERS[id_]
                    except KeyError:
                        try:
                            user = await command_context.client.user_get(id_)
                        except BaseException as err:
                            if not (isinstance(err, ConnectionError) or
                                isinstance(err, DiscordException) and err.code in (
                                    ERROR_CODES.unknown_user,
                                    ERROR_CODES.unknown_member,
                                        )):
                                    raise
                        else:
                            return user
                    else:
                        return user
                
                else:
                    channel = message.channel
                    guild = message.guild
                    if guild is None:
                        if (not isinstance(channel, ChannelGuildBase)):
                            for user in channel.users:
                                if user.id == id_:
                                    return user
                    
                    else:
                        try:
                            user = guild.users[id_]
                        except KeyError:
                            pass
                        else:
                            return user
        
        if flags&CONVERTER_FLAG_MENTION:
            user = parse_user_mention(part, message)
            if (user is not None):
                return user
        
        if flags&CONVERTER_FLAG_NAME:
            channel = message.channel
            guild = channel.guild
            if (guild is None):
                if isinstance(channel, ChannelGuildBase):
                    user = None
                else:
                    user = channel.get_user_like(part)
            else:
                user = guild.get_user_like(part)
            
            if (user is not None):
                return user
        
        return None
else:
    async def user_converter(command_context, content_parser_parameter_detail, part):
        flags = content_parser_parameter_detail.flags
        message = command_context.message
        
        if flags&CONVERTER_FLAG_ID:
            parsed = ID_RP.fullmatch(part)
            if (parsed is not None):
                id_ = int(parsed.group(1))
                
                if flags&CONVERTER_FLAG_EVERYWHERE:
                    if flags&CONVERTER_FLAG_PROFILE:
                        guild = message.guild
                        if (guild is not None):
                            try:
                                user = await command_context.client.guild_user_get(guild, id_)
                            except BaseException as err:
                                if not (isinstance(err, ConnectionError) or
                                    isinstance(err, DiscordException) and err.code in (
                                        ERROR_CODES.unknown_user,
                                        ERROR_CODES.unknown_member,
                                            )):
                                        raise
                            else:
                                return user
                    
                    try:
                        user = await command_context.client.user_get(id_)
                    except BaseException as err:
                        if not (isinstance(err, ConnectionError) or
                            isinstance(err, DiscordException) and err.code in (
                                ERROR_CODES.unknown_user,
                                ERROR_CODES.unknown_member,
                                    )):
                                raise
                    else:
                        return user
                
                else:
                    channel = message.channel
                    guild = message.guild
                    if guild is None:
                        if (not isinstance(channel, ChannelGuildBase)):
                            for user in channel.users:
                                if user.id == id_:
                                    return user
                    
                    else:
                        try:
                            user = await command_context.client.guild_user_get(guild, id_)
                        except BaseException as err:
                            if not (isinstance(err, ConnectionError) or
                                isinstance(err, DiscordException) and err.code in (
                                    ERROR_CODES.unknown_user,
                                    ERROR_CODES.unknown_member,
                                        )):
                                    raise
                        else:
                            return user
        
        if flags&CONVERTER_FLAG_MENTION:
            user = parse_user_mention(part, message)
            if (user is not None):
                return user
        
        if flags&CONVERTER_FLAG_NAME:
            channel = message.channel
            guild = channel.guild
            if (guild is None):
                if (not isinstance(channel, ChannelGuildBase)):
                    user = channel.get_user_like(part)
                    if (user is not None):
                        return user
            
            else:
                try:
                    user = await command_context.client.guild_user_search(guild, part)
                except BaseException as err:
                    if not (isinstance(err, ConnectionError) or
                        isinstance(err, DiscordException) and err.code in (
                            ERROR_CODES.unknown_user,
                            ERROR_CODES.unknown_member,
                                )):
                            raise
                else:
                    return user
        
        return None

CONVERTER_USER = ConverterSetting(
    converter = user_converter,
    uses_flags = True,
    default_flags = ConverterFlag.user_default,
    all_flags = ConverterFlag.user_all,
    alternative_type_name = 'user',
    default_type = User,
    alternative_types = [
        UserBase,
    ],
    requires_part = True,
)

async def client_converter(command_context, content_parser_parameter_detail, part):
    flags = content_parser_parameter_detail.flags
    message = command_context.message
    
    if flags&CONVERTER_FLAG_ID:
        parsed = ID_RP.fullmatch(part)
        if (parsed is not None):
            id_ = int(parsed.group(1))
            
            try:
                client = CLIENTS[id_]
            except KeyError:
                pass
            else:
                if flags&CONVERTER_FLAG_EVERYWHERE:
                    return client
                
                else:
                    if client in message.channel.clients:
                        return client
    
    if flags&CONVERTER_FLAG_MENTION:
        client = parse_user_mention(part, message)
        if (client is not None) and isinstance(client, Client):
            return client
    
    if flags&CONVERTER_FLAG_NAME:
        if flags&CONVERTER_FLAG_EVERYWHERE:
            clients = list(CLIENTS.values())
        else:
            clients = message.channel.clients
        
        if 1 < len(part) < 38:
            if len(part) > 6 and part[-5] == '#':
                try:
                    discriminator = int(part[-4:])
                except ValueError:
                    pass
                else:
                    name_ = part[:-5]
                    for client in clients:
                        if (client.discriminator == discriminator) and (client.name == name_):
                            return client
            
            if len(part) < 32:
                pattern = re.compile(re.escape(part), re.I)
                for client in clients:
                    if (pattern.match(client.name) is not None):
                        return client
                
                guild = message.guild
                if (guild is not None):
                    for client in clients:
                        try:
                            guild_profile = client.guild_profiles[guild.id]
                        except KeyError:
                            continue
                        
                        nick = guild_profile.nick
                        
                        if nick is None:
                            continue
                        
                        if pattern.match(nick) is None:
                            continue
                        
                        return client
    
    return None
    
CONVERTER_CLIENT = ConverterSetting(
    converter = client_converter,
    uses_flags = True,
    default_flags = ConverterFlag.client_default,
    all_flags = ConverterFlag.client_all,
    alternative_type_name = 'client',
    default_type = Client,
    alternative_types = None,
    requires_part = True,
)

async def channel_converter(command_context, content_parser_parameter_detail, part):
    flags = content_parser_parameter_detail.flags
    channel_type = content_parser_parameter_detail.type
    message = command_context.message
    
    if flags&CONVERTER_FLAG_ID:
        parsed = ID_RP.fullmatch(part)
        if (parsed is not None):
            id_ = int(parsed.group(1))
            
            if flags&CONVERTER_FLAG_EVERYWHERE:
                try:
                    channel = CHANNELS[id_]
                except KeyError:
                    pass
                else:
                    if ((channel_type is None) or isinstance(channel, channel_type)):
                        return channel
            
            else:
                channel = message.channel
                guild = message.guild
                if guild is None:
                    if ((channel_type is None) or isinstance(channel, channel_type)) and channel.id == id_:
                        return channel
                
                else:
                    try:
                        channel = guild.channels[id_]
                    except KeyError:
                        pass
                    else:
                        if ((channel_type is None) or isinstance(channel, channel_type)):
                            return channel
    
    if flags&CONVERTER_FLAG_MENTION:
        channel = parse_channel_mention(part, message)
        if (channel is not None):
            if ((channel_type is None) or isinstance(channel, channel_type)):
                return channel
    
    if flags&CONVERTER_FLAG_NAME:
        channel = message.channel
        guild = channel.guild
        if guild is None:
            if ((channel_type is None) or isinstance(channel, channel_type)) and channel.has_name_like(part):
                return channel
        else:
            channel = guild.get_channel_like(part, type_=channel_type)
            if (channel is not None):
                return channel
    
    return None

CONVERTER_CHANNEL = ConverterSetting(
    converter = channel_converter,
    uses_flags = True,
    default_flags = ConverterFlag.channel_default,
    all_flags = ConverterFlag.channel_all,
    alternative_type_name = 'channel',
    default_type = ChannelBase,
    alternative_types = [
        ChannelGuildBase,
        ChannelTextBase,
        ChannelVoiceBase,
        ChannelText,
        ChannelPrivate,
        ChannelVoice,
        ChannelGroup,
        ChannelCategory,
        ChannelStore,
        ChannelThread,
        ChannelStage,
        ChannelDirectory,
    ],
    requires_part = True,
)

async def role_converter(command_context, content_parser_parameter_detail, part):
    flags = content_parser_parameter_detail.flags
    message = command_context.message
    
    if flags&CONVERTER_FLAG_ID:
        parsed = ID_RP.fullmatch(part)
        if (parsed is not None):
            id_ = int(parsed.group(1))
            
            if flags&CONVERTER_FLAG_EVERYWHERE:
                try:
                    role = ROLES[id_]
                except KeyError:
                    pass
                else:
                    return role
            
            else:
                guild = message.guild
                if (guild is not None):
                    try:
                        role = guild.roles[id_]
                    except KeyError:
                        pass
                    else:
                        return role
    
    if flags&CONVERTER_FLAG_MENTION:
        role = parse_role_mention(part)
        if (role is not None):
            return role
    
    if flags&CONVERTER_FLAG_NAME:
        guild = message.guild
        if (guild is not None):
            role = guild.get_role_like(part)
            if (role is not None):
                return role
    
    return None

CONVERTER_ROLE = ConverterSetting(
    converter = role_converter,
    uses_flags = True,
    default_flags = ConverterFlag.role_default,
    all_flags = ConverterFlag.role_all,
    alternative_type_name = 'role',
    default_type = Role,
    alternative_types = None,
    requires_part = True,
)

async def emoji_converter(command_context, content_parser_parameter_detail, part):
    flags = content_parser_parameter_detail.flags
    if flags&CONVERTER_FLAG_MENTION:
        emoji = parse_emoji(part)
        if (emoji is not None):
            return emoji
    
    message = command_context.message
    if flags&CONVERTER_FLAG_ID:
        parsed = ID_RP.fullmatch(part)
        if (parsed is not None):
            id_ = int(parsed.group(1))
            
            if flags&CONVERTER_FLAG_EVERYWHERE:
                try:
                    emoji = EMOJIS[id_]
                except KeyError:
                    pass
                else:
                    return emoji
            
            else:
                guild = message.guild
                if (guild is not None):
                    try:
                        emoji = guild.emojis[id_]
                    except KeyError:
                        pass
                    else:
                        return emoji
    
    if flags&CONVERTER_FLAG_NAME:
        guild = message.guild
        if (guild is not None):
            emoji = guild.get_emoji_like(part)
            if (emoji is not None):
                return emoji
    
    return None

CONVERTER_EMOJI = ConverterSetting(
    converter = emoji_converter,
    uses_flags = True,
    default_flags = ConverterFlag.emoji_default,
    all_flags = ConverterFlag.emoji_all,
    alternative_type_name = 'emoji',
    default_type = Emoji,
    alternative_types = None,
    requires_part = True,
)


async def sticker_converter(command_context, content_parser_parameter_detail, part):
    flags = content_parser_parameter_detail.flags
    
    message = command_context.message
    if flags&CONVERTER_FLAG_ID:
        parsed = ID_RP.fullmatch(part)
        if (parsed is not None):
            id_ = int(parsed.group(1))
            
            if flags&CONVERTER_FLAG_EVERYWHERE:
                try:
                    sticker = await command_context.client.sticker_get(id_)
                except BaseException as err:
                    if not (isinstance(err, ConnectionError) or
                        isinstance(err, DiscordException) and (err.code == ERROR_CODES.unknown_sticker)
                                ):
                            raise
                else:
                    return sticker
            
            else:
                guild = message.guild
                if (guild is not None):
                    try:
                        sticker = await command_context.client.sticker_get(id_)
                    except BaseException as err:
                        if not (isinstance(err, ConnectionError) or
                            isinstance(err, DiscordException) and (err.code == ERROR_CODES.unknown_sticker)
                                    ):
                                raise
                    else:
                        return sticker
    
    if flags&CONVERTER_FLAG_NAME:
        guild = message.guild
        if (guild is not None):
            sticker = guild.get_sticker_like(part)
            if (sticker is not None):
                return sticker
    
    return None

STICKER_EMOJI = ConverterSetting(
    converter = sticker_converter,
    uses_flags = True,
    default_flags = ConverterFlag.sticker_default,
    all_flags = ConverterFlag.sticker_all,
    alternative_type_name = 'sticker',
    default_type = Sticker,
    alternative_types = None,
    requires_part = True,
)


async def guild_converter(command_context, part, content_parser_parameter_detail):
    parsed = ID_RP.fullmatch(part)
    if (parsed is None):
        return None
    
    id_ = int(parsed.group(1))
    
    try:
        guild = GUILDS[id_]
    except KeyError:
        return None
    
    if content_parser_parameter_detail.flags&CONVERTER_FLAG_EVERYWHERE:
        return guild
    
    if guild in command_context.client.guild_profiles:
        return guild
    
    return None

CONVERTER_GUILD = ConverterSetting(
    converter = guild_converter,
    uses_flags = True,
    default_flags = ConverterFlag.guild_default,
    all_flags = ConverterFlag.guild_all,
    alternative_type_name = 'guild',
    default_type = Guild,
    alternative_types = None,
    requires_part = True
)

# Gets a message by it's id
async def _message_converter_m_id(command_context, content_parser_parameter_detail, message_id):
    message = MESSAGES.get(message_id, None)
    channel = command_context.message.channel
    if (message is not None):
        # Message found
        if content_parser_parameter_detail.flags&CONVERTER_FLAG_EVERYWHERE:
            return message
        else:
            # Only local message can be yielded, so check if it is local
            guild = channel.guild
            if guild is None:
                if message.channel is channel:
                    return message
                else:
                    # Message found, but other channel, yield None
                    return None
            else:
                if message.guild is guild:
                    return message
                else:
                    # Message found, but other guild, yield None
                    return None
    
    # Try to get message by id
    client = command_context.client
    if channel.cached_permissions_for(client).can_read_message_history:
        try:
            message = await client.message_get(channel, message_id)
        except BaseException as err:
            if not (isinstance(err, ConnectionError) or
                isinstance(err, DiscordException) and err.code in (
                    ERROR_CODES.unknown_channel, # message deleted
                    ERROR_CODES.unknown_message, # channel deleted
                    ERROR_CODES.missing_access, # client removed
                    ERROR_CODES.missing_permissions, # permissions changed meanwhile
                        )):
                    raise
            
            # Message do not exists at the respective channel, or any other acceptable error
            return None
        else:
            return message
    else:
        # The message is given by id, but the client request it.
        return None

# Gets a message by it's and it's channel's id
async def _message_converter_cm_id(command_context, content_parser_parameter_detail, channel_id, message_id):
    channel = command_context.message.channel
    message = MESSAGES.get(message_id, None)
    if (message is not None):
        # Message found
        if content_parser_parameter_detail.flags&CONVERTER_FLAG_EVERYWHERE:
            return message
        else:
            # Only local message can be yielded, so check if it is local
            guild = channel.guild
            if (message.channel is channel) if (guild is None) else (message.guild is guild):
                return message
        
        # Message found, but other guild or channel yield None
        return None
    
    message_channel = CHANNELS.get(channel_id, None)
    if (message_channel is None):
        return None

    if content_parser_parameter_detail.flags&CONVERTER_FLAG_EVERYWHERE:
        # Lets use that multi client core
        for client in message_channel.clients:
            if message_channel.cached_permissions_for(client).can_read_message_history:
                try:
                    message = await client.message_get(message_channel,  message_id)
                except BaseException as err:
                    if isinstance(err, ConnectionError):
                        return None
                    
                    if isinstance(err, DiscordException):
                        err_code = err.code
                        # If the message or channel is deleted, return None
                        if err_code in (
                            ERROR_CODES.unknown_channel, # message deleted
                            ERROR_CODES.unknown_message, # channel deleted
                                ):
                            return None
                        
                        # If client is removed or has it's permissions changed, lets move on the next if applicable
                        if err_code in (
                            ERROR_CODES.missing_access, # client removed
                            ERROR_CODES.missing_permissions, # permissions changed meanwhile
                                ):
                            continue
                    
                    raise
                else:
                    return message
        
        # No message could be requested successfully.
        return None
    
    guild = channel.guild
    if (message_channel is channel) if (guild is None) else (message_channel.guild is guild):
        client = command_context.client
        if channel.cached_permissions_for(client).can_read_message_history:
            try:
                message = await client.message_get(message_channel, message_id)
            except BaseException as err:
                if not (isinstance(err, ConnectionError) or
                    isinstance(err, DiscordException) and err.code in (
                        ERROR_CODES.unknown_channel, # message deleted
                        ERROR_CODES.unknown_message, # channel deleted
                        ERROR_CODES.missing_access, # client removed
                        ERROR_CODES.missing_permissions, # permissions changed meanwhile
                            )):
                        raise
            else:
                return message
        
        return None

async def message_converter(command_context, content_parser_parameter_detail, part):
    if content_parser_parameter_detail.flags&CONVERTER_FLAG_ID:
        parsed = ID_RP.fullmatch(part)
        if (parsed is not None):
            message_id = int(parsed.group(1))
            return await _message_converter_m_id(command_context, content_parser_parameter_detail, message_id)
        
        parsed = CHANNEL_MESSAGE_RP.fullmatch(part)
        if (parsed is not None):
            channel_id, message_id = parsed.groups()
            channel_id = int(channel_id)
            message_id = int(message_id)
            return await _message_converter_cm_id(command_context, content_parser_parameter_detail, channel_id,
                message_id)
    
    if content_parser_parameter_detail.flags&CONVERTER_FLAG_URL:
        parsed = MESSAGE_JUMP_URL_RP.fullmatch(part)
        if (parsed is not None):
            _, channel_id, message_id = parsed.groups()
            channel_id = int(channel_id)
            message_id = int(message_id)
            return await _message_converter_cm_id(command_context, content_parser_parameter_detail, channel_id,
                message_id)
    
    return None

CONVERTER_MESSAGE = ConverterSetting(
    converter = message_converter,
    uses_flags = True,
    default_flags = ConverterFlag.message_default,
    all_flags = ConverterFlag.message_all,
    alternative_type_name = 'message',
    default_type = Message,
    alternative_types = None,
    requires_part = True,
)

async def invite_converter(command_context, part, content_parser_parameter_detail):
    flags = content_parser_parameter_detail.flags
    
    # It would not be a Huyane code without some GOTO
    while True:
        if flags&CONVERTER_FLAG_URL:
            parsed = INVITE_URL_RP.fullmatch(part)
            if parsed is not None:
                break
        
        if flags&CONVERTER_FLAG_ID:
            parsed = INVITE_CODE_RP.fullmatch(part)
            if (parsed is not None):
                break
        
        return None
    
    code = parsed.group(1)
    
    try:
        invite = await command_context.client.invite_get(code)
    except BaseException as err:
        if not (isinstance(err, ConnectionError) or
            isinstance(err, DiscordException) and err.code == ERROR_CODES.unknown_invite # Invite not exists
                    ):
                raise
        
        return None
    
    return invite

CONVERTER_INVITE = ConverterSetting(
    converter = invite_converter,
    uses_flags = True,
    default_flags = ConverterFlag.invite_default,
    all_flags = ConverterFlag.invite_all,
    alternative_type_name = 'invite',
    default_type = Invite,
    alternative_types = None,
    requires_part = True,
)


async def color_converter(command_context, content_parser_parameter_detail, part):
    return parse_color(part)

CONVERTER_COLOR = ConverterSetting(
    converter = color_converter,
    uses_flags = False,
    default_flags = ConverterFlag(),
    all_flags = ConverterFlag(),
    alternative_type_name = 'color',
    default_type = Color,
    alternative_types = None,
    requires_part = True,
)


async def str_converter(command_context, content_parser_parameter_detail, part):
    return part

CONVERTER_STR = ConverterSetting(
    converter = str_converter,
    uses_flags = False,
    default_flags = ConverterFlag(),
    all_flags = ConverterFlag(),
    alternative_type_name = None,
    default_type = str,
    alternative_types = None,
    requires_part = True,
)


async def int_converter(command_context, content_parser_parameter_detail, part):
    if len(part) > NUMERIC_CONVERSION_LIMIT:
        return None
    
    try:
        int_ = int(part)
    except ValueError:
        int_ = None
    
    return int_

CONVERTER_INT = ConverterSetting(
    converter = int_converter,
    uses_flags = False,
    default_flags = ConverterFlag(),
    all_flags = ConverterFlag(),
    alternative_type_name = None,
    default_type = int,
    alternative_types = None,
    requires_part = True,
)

async def tdelta_converter(command_context, content_parser_parameter_detail, part):
    return parse_tdelta(part)

CONVERTER_TDELTA = ConverterSetting(
    converter = tdelta_converter,
    uses_flags = False,
    default_flags = ConverterFlag(),
    all_flags = ConverterFlag(),
    alternative_type_name = 'tdelta',
    default_type = timedelta,
    alternative_types = None,
    requires_part = True,
)

if (relativedelta is not None):
    async def rdelta_converter(command_context, content_parser_parameter_detail, part):
        return parse_rdelta(part)

    CONVERTER_RDELTA = ConverterSetting(
        converter = rdelta_converter,
        uses_flags = False,
        default_flags = ConverterFlag(),
        all_flags = ConverterFlag(),
        alternative_type_name = 'rdelta',
        default_type = relativedelta,
        alternative_types = None,
        requires_part = True,
    )

else:
    rdelta_converter = None
    CONVERTER_RDELTA = None


class ContentParserParameterDetail:
    """
    Stores details about a converter.
    
    Attributes
    ----------
    converter_setting : ``ConverterSetting``
        The converter setting used by the parameter.
    flags : ``ConverterFlag``
        Converter flags to customize the events.
    type : `None` or `type` instance
        The type or subtype of the annotation to parse.
    """
    __slots__ = ('converter_setting', 'flags', 'type',)
    def __new__(cls, converter_setting, type_):
        """
        Creates a new ``ContentParserParameterDetail`` instance with the given parameters.
        
        Parameters
        ----------
        converter_setting : ``ConverterSetting``
            The converter setting used by the parameter.
        type_ : `None` or `type` instance
            The type or subtype of the annotation to parse.
        """
        self = object.__new__(cls)
        self.type = type_
        self.converter_setting = converter_setting
        self.flags = converter_setting.default_flags
        return self
    
    def __repr__(self):
        """Returns the ``ContentParserParameterDetail``'s representation."""
        result = ['<', self.__class__.__name__, ' converter_setting=']
        converter_setting = self.converter_setting
        result.append(repr(converter_setting))
        
        type_ = self.type
        if (type_ is not converter_setting.default_type):
            result.append(', type=')
            result.append(repr(type_))
        
        flags = self.flags
        if (flags != converter_setting.default_flags):
            result.append(', flags=')
            result.append(repr(flags))
        
        result.append('>')
        return ''.join(result)


class ContentParserParameter:
    """
    Represents a parameter to parse.
    
    Parameters
    ----------
    default : `None` or `Any`
        Default value to the parser
    has_default : `bool`
        Whether the parser has default value.
    default : `None` or `Any`
        The default object to return if the parser fails.
    description : `None` or `str`
        The description of the parameter if any.
    detail : `None` or ``ContentParserParameterDetail``
        Converting details for single-type annotation.
    details : `None` or `list` of ``ContentParserParameterDetail``
        Converting details for multi-type annotations.
    display_name : `str`
        The parameter's display name.
    has_default : `bool`
        Whether the parameter has default value set.
    index : `int`
        The parameter's index.
    is_args : `bool`
        Whether the parameter is `*args` parameter.
    is_keyword : `bool`
        Whether the parameter is keyword only parameter.
    is_kwargs : `bool`
        Whether the parameter is `**kwargs` parameter.
    is_positional : `bool`
        Whether the parameter is positional.
    is_rest : `bool`
        Whether the parameter is rest parser.
    name : `str`
        The parameter's name.
    """
    __slots__ = ('converter', 'default', 'description', 'detail', 'details', 'display_name', 'flags', 'has_default',
        'index', 'is_args', 'is_keyword', 'is_kwargs', 'is_positional', 'is_rest', 'name')
    
    def __new__(cls, parameter, index):
        """
        Creates a new ``ContentParserParameter`` with the given parameters.
        
        Parameters
        ----------
        parameter : ``Parameter``
            The analyzed parameter to process.
        index : `int`
            The parameter's index.
        
        Raises
        ------
        TypeError
            - If `annotation`'s type is invalid.
            - If `description`'s type is invalid.
            - If `name`'s type is invalid.
        ValueError
            - There is no converter for the given `annotation`.
            - The `annotation` is a `tuple`, but it's length is not 2 or 3.
        RuntimeError
            - Multi-type annotation without requiring parsing is forbidden.
        """
        display_name = parameter.name
        
        details = []
        
        while True:
            if not parameter.has_annotation:
                description = None
                break
            
            annotation = parameter.annotation
            if annotation is None:
                description = None
                break
            
            detail = get_detail_for_value(annotation)
            if (detail is not None):
                details.append(detail)
                description = None
                break
            
            if isinstance(annotation, set):
                details.extend(get_details_from_set(annotation))
                description = None
                break
            
            if isinstance(annotation, tuple):
                if type(annotation) is not tuple:
                    annotation = tuple(annotation)
                
                annotation_length = len(annotation)
                if annotation_length not in (2, 3):
                    raise ValueError(f'`tuple` annotation\'s length can be either `2` or `3`, got '
                        f'{annotation_length!r}: {annotation!r}')
                
                annotation_tuple_type = annotation[0]
                if (annotation_tuple_type is not None):
                    detail = get_detail_for_value(annotation_tuple_type)
                    if (detail is not None):
                        details.append(detail)
                    else:
                        if isinstance(annotation_tuple_type, set):
                            details.extend(get_details_from_set(annotation_tuple_type))
                        else:
                            raise TypeError(f'`annotation` was given as `tuple`, but it\'s 0th element was not given '
                                f'as any of the expected values: `None`, `type`, `str`, `set`, instance, got '
                                f'{annotation_tuple_type.__class__.__name__}; {annotation_tuple_type!r}.')
                
                
                annotation_tuple_description = annotation[1]
                if type(annotation_tuple_description) is str:
                    description = annotation_tuple_description
                elif isinstance(annotation_tuple_description, str):
                    description = str(annotation_tuple_description)
                else:
                    raise TypeError(f'`annotation` description can be `str` instance, got '
                        f'{annotation_tuple_description.__class__.__name__}.')
                
                
                if annotation_length == 3:
                    annotation_tuple_name = annotation[2]
                    if type(annotation_tuple_name) is str:
                        display_name = annotation_tuple_name
                    elif isinstance(annotation_tuple_name, str):
                        display_name = str(annotation_tuple_name)
                    else:
                        raise TypeError(f'`annotation` name can be `str` instance, got '
                            f'{annotation_tuple_name.__class__.__name__}.')
                
                break
            
            raise TypeError(f'`annotation` can be either given as `None`, `type`, `str`, `tuple, or `set`, got '
                f'{annotation.__class__.__name__}; {annotation!r}.')
        
        
        details_length = len(details)
        if details_length == 0:
            detail = ContentParserParameterDetail(CONVERTER_NONE, CONVERTER_NONE.default_type)
            details = None
        elif details_length == 1:
            detail = details[0]
            details = None
        else:
            for detail in details:
                if not detail.converter_setting.requires_part:
                    raise RuntimeError('Multi-type annotation without requiring parsing is forbidden.')
            detail = None
        
        display_name = raw_name_to_display(display_name)
        
        has_default = parameter.has_default
        if has_default:
            default = parameter.default
        else:
            default = None
        
        is_positional = parameter.is_positional()
        is_keyword = parameter.is_keyword_only()
        is_args = parameter.is_args()
        is_kwargs = parameter.is_kwargs()
        
        name = parameter.name
        
        self = object.__new__(cls)
        self.detail = detail
        self.details = details
        self.default = default
        self.description = description
        self.display_name = display_name
        self.has_default = has_default
        self.index = index
        self.is_args = is_args
        self.is_rest = False
        self.is_keyword = is_keyword
        self.is_kwargs = is_kwargs
        self.is_positional = is_positional
        self.name = name
        return self
    
    def __repr__(self):
        """Returns the inter's representation."""
        repr_parts = ['<', self.__class__.__name__,
            ' name=', repr(self.name),
         ]
        
        if self.has_default:
            repr_parts.append(', default=')
            repr_parts.append(repr(self.default))
        
        detail = self.detail
        if (detail is None):
            repr_parts.append(', details=')
            repr_parts.append(repr(self.details))
        else:
            if (detail.converter_setting is CONVERTER_NONE):
                if self.is_rest:
                    repr_parts.append(', is_rest=True')
            else:
                repr_parts.append(', details=[')
                repr_parts.append(repr(detail))
                repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    def set_converter_setting(self, converter_setting):
        """
        Sets a new converter setting to the ``ContentParserParameter``.
        
        Parameters
        ----------
        converter_setting : ``ConverterSetting``
        
        Raises
        ------
        RuntimeError
            Converter setting cannot be set if the parser is multi type events.
        """
        if self.detail is None:
            raise RuntimeError('Converter setting cannot be set if the parser is multi type events.')
        
        self.detail = ContentParserParameterDetail(converter_setting, converter_setting.default_type)

    async def parse(self, command_context, part):
        """
        Tries to parse the parameter from the given part.
        
        This method is a coroutine.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            The respective command's context.
        part : `None` or `str`
            The received content's part to check.
        
        Returns
        -------
        parsed : `None` or `Any`
            The parsed object if any.
        """
        detail = self.detail
        if (detail is None):
            for detail in self.details:
                parsed = await detail.converter_setting.converter(command_context, detail, part)
                if (parsed is not None):
                    break
            else:
                parsed = None
        else:
            converter_setting = detail.converter_setting
            converter = converter_setting.converter
            if converter_setting.requires_part:
                parsed = await converter(command_context, detail, part)
            else:
                parsed = await converter(command_context, detail)
        
        return parsed
    
    
    def _iter_details(self):
        """
        Iterates over the details of the converter.
        
        This method is a generator.
        
        Yields
        ------
        detail : ``ContentParserParameterDetail``
        """
        detail = self.detail
        if (detail is None):
            yield from self.details
        else:
            yield self.detail


COMMAND_CONTENT_PARSER_POST_PROCESSORS = []


def get_detail_for_str(annotation):
    """
    Creates a ``ContentParserParameterDetail`` for the given string annotation value.
    
    Parameters
    ----------
    annotation : `str`
        The respective annotation.

    Returns
    -------
    detail : ``ContentParserParameterDetail``
    
    Raises
    ------
    ValueError
        There is no converter setting for the annotation.
    """
    try:
        converter_setting = CONVERTER_SETTING_NAME_TO_SETTING[annotation]
    except KeyError:
        raise ValueError(f'There is no converter registered for {annotation!r}.') from None
    
    annotation_type = CONVERTER_NAME_TO_TYPE.get(annotation, None)
    
    return ContentParserParameterDetail(converter_setting, annotation_type)


def get_detail_for_type(annotation):
    """
    Creates a ``ContentParserParameterDetail`` for the given type annotation value.
    
    Parameters
    ----------
    annotation : `str`
        The respective annotation.
    
    Returns
    -------
    detail : ``ContentParserParameterDetail``
    
    Raises
    ------
    ValueError
        There is no converter setting for the annotation.
    """
    try:
        converter_setting = CONVERTER_SETTING_TYPE_TO_SETTING[annotation]
    except KeyError:
        raise ValueError(f'There is no converter registered for {annotation!r}.') from None
    
    return ContentParserParameterDetail(converter_setting, annotation)



def get_detail_for_value(annotation):
    """
    Creates a ``ContentParserParameterDetail`` for the given value.
    
    Parameters
    ----------
    annotation : `str`, `type`
        The respective annotation.
    
    Returns
    -------
    detail : ``ContentParserParameterDetail``
    
    Raises
    ------
    ValueError
        The is no converter setting for the annotation.
    """
    if isinstance(annotation, type):
        detail = get_detail_for_type(annotation)
    
    elif type(annotation) is str:
        detail = get_detail_for_str(annotation)
    
    elif isinstance(annotation, str):
        annotation = str(annotation)
        detail = get_detail_for_str(annotation)
    
    else:
        detail = None
    
    return detail

def get_details_from_set(annotation):
    """
    Generates ``ContentParserParameterDetail`` from the given annotation value.
    
    This function is a generator.
    
    Parameters
    ----------
    annotation : `set` of (`str`, `type`)
        The respective annotation.
    
    Yields
    -------
    detail : ``ContentParserParameterDetail``
    
    Raises
    ------
    ValueError
        The is no converter setting for the annotation.
    TypeError
        If a sub-annotation's type is incorrect.
    """
    for sub_annotation in annotation:
        if sub_annotation is None:
            continue
        
        detail = get_detail_for_value(sub_annotation)
        if (detail is None):
            raise TypeError(f'`annotation` set can contain only `str` and `type` elements, got '
                f'{sub_annotation.__class__.__name__}.')
        
        yield detail
        continue


class CommandContentParser:
    """
    Content parser for commands.
    
    Attributes
    ----------
    _parameters : `list` of ``ContentParserParameter``
        The parameters of the respective function.
    _content_parameter_parser : ``ContentParameterParser``
        The parameter separator of the parser.
    """
    __slots__ = ('_content_parameter_parser', '_parameters',)
    
    def __new__(cls, func, separator, assigner):
        """
        Creates a new ``CommandContentParser`` instance returning the parser for the function and the function itself
        as well.
        
        Parameters
        ----------
        func : `async-callable`
            The callable function.
        separator : `None`, ``ContentParameterSeparator``, `str` or `tuple` (`str`, `str`)
            The parameter separator of the parser.
        
        Returns
        -------
        self : ``CommandContentParser``
            The created parser.
        func : `async-callable`
            The function to which the parser is created for.
        
        Raises
        ------
        TypeError
            - If `separator` is not given as `None`, `str`, neither as `tuple` instance.
            - If `separator` was given as `tuple`, but it's element are not `str` instances.
            - If `assigner` was not given neither as `None` or as `str` instance.
            - If an `annotation`'s type is invalid.
            - If a `description`'s type is invalid.
            - If a `name`'s type is invalid.
            - If `func` is not `async-callable` and cannot instanced into `async-callable` either.
        ValueError
            - If `separator` is given as `str`, but it's length is not `1`.
            - If `separator` is given as `str`, but it is a space character.
            - If `separator` is given as `tuple`, but one of it's element's length is not `1`.
            - If `separator` is given as `tuple`, but one of it's element's is a space character.
            - If `assigner`'s length is not `1`.
            - There is no converter for a given `annotation`.
            - The an `annotation` is a `tuple`, but it's length is not 2 or 3.
        """
        content_parameter_parser = ContentParameterParser(separator, assigner)
        
        analyzer = CallableAnalyzer(func)
        if analyzer.is_async() or analyzer.is_async_generator():
            real_analyzer = analyzer
            should_instance = False
        
        elif analyzer.can_instance_to_async_callable() or analyzer.can_instance_to_async_generator():
            real_analyzer = CallableAnalyzer(func.__call__, as_method=True)
            if (not real_analyzer.is_async()) and (not real_analyzer.is_async_generator()):
                raise TypeError(f'`func` is not `async-callable` and cannot be instanced to `async` either, got '
                    f'{func!r}.')
            
            should_instance = True
        
        else:
            raise TypeError(f'`func` is not `async-callable` and cannot be instanced to `async` either, got {func!r}.')
        
        parameters = []
        
        index = 0
        for parameter in real_analyzer.parameters:
            if not parameter.reserved:
                content_parser_parameter = ContentParserParameter(parameter, index)
                parameters.append(content_parser_parameter)
                index += 1
        
        if should_instance:
            func = analyzer.insatnce()
        
        self = object.__new__(cls)
        self._parameters = parameters
        self._content_parameter_parser = content_parameter_parser
        
        for postprocessor in CONTENT_PARSER_PARAMETER_POSTPROCESSORS:
            postprocessor(self)
        
        return self, func
    
    
    async def parse_content(self, command_context, index):
        """
        Parses the message's content to fulfill a command's parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        command_context : ``CommandContext``
            A command context of the command.
        index : `int`
            The index of the content's character since the command's parameters should be parsed from.
        
        Returns
        -------
        parameter_parsing_states : `list` of ``ParameterParsingStateBase`` instances
        """
        content_parameter_parser = self._content_parameter_parser
        parameter_parsing_states = create_parameter_parsing_states(self)
        
        for parameter_parsing_state in iter_no_part_parameter_states(parameter_parsing_states):
            parsed_value = await parameter_parsing_state.content_parser_parameter.parse(command_context, None)
            parameter_parsing_state.add_parsed_value(parsed_value, None)
        
        content = command_context.content
        length = len(content)
        while True:
            if index >= length:
                break
            
            if get_is_all_parameter_parser_state_satisfied(parameter_parsing_states):
                parameter_parsing_state = get_rest_parameter_state(parameter_parsing_states)
                if (parameter_parsing_state is not None):
                    rest = parse_rest_content(content, index)
                    parameter_parsing_state.add_parsed_value(rest, None)
                
                break
            
            keyword, part, index = content_parameter_parser(content, index)
            if keyword is None:
                parameter_parsing_state = get_next_non_filled_parameter_state(parameter_parsing_states)
                if parameter_parsing_state is None:
                    continue
                
                parsed_value = await parameter_parsing_state.content_parser_parameter.parse(command_context, part)
                parameter_parsing_state.add_parsed_value(parsed_value, None)
            else:
                keyword = raw_name_to_display(keyword)
                parameter_parsing_state = get_keyword_parameter_state(parameter_parsing_states, keyword)
                if parameter_parsing_state is None:
                    continue
                
                parsed_value = await parameter_parsing_state.content_parser_parameter.parse(command_context, part)
                parameter_parsing_state.add_parsed_value(parsed_value, keyword)
        
        return parameter_parsing_states


def get_next_non_filled_parameter_state(parameter_parsing_states):
    """
    Gets the next non-filled parameter parsing state.
    
    Parameters
    ----------
    parameter_parsing_states : `list` of ``ParameterParsingState``
        Parameters parsing states to choose from.
    
    Returns
    -------
    parameter_parsing_state : ``ParameterParsingState`` or `None`
    """
    for parameter_parsing_state in parameter_parsing_states:
        content_parser_parameter = parameter_parsing_state.content_parser_parameter
        if content_parser_parameter.is_keyword or content_parser_parameter.is_kwargs:
            return None
        
        if content_parser_parameter.is_args:
            return parameter_parsing_state
        
        if not parameter_parsing_state.is_satisfied():
            return parameter_parsing_state
        
        continue
    
    return None


def iter_no_part_parameter_states(parameter_parsing_states):
    """
    Iterates over the parameter parsing states, and yields which do not require
    
    Parameters
    ----------
    parameter_parsing_states : `list` of ``ParameterParsingState``
        Parameters parsing states to choose from.

    Yields
    ------
    parameter_parsing_state : ``ParameterParsingState``
    """
    for parameter_parsing_state in parameter_parsing_states:
        content_parser_parameter = parameter_parsing_state.content_parser_parameter
        if content_parser_parameter.is_rest:
            continue
        
        detail = content_parser_parameter.detail
        if detail is None:
            continue
        
        if detail.converter_setting.requires_part:
            continue
        
        yield parameter_parsing_state


def get_keyword_parameter_state(parameter_parsing_states, keyword):
    """
    Tries to get the parameter state for the given name.
    
    Parameters
    ----------
    parameter_parsing_states : `list` of ``ParameterParsingState``
        Parameters parsing states to choose from.
    keyword : `str`
        The keyword to get the parameter for.
    
    Returns
    -------
    parameter_parsing_state : ``ParameterParsingState`` or `None`
    """
    for parameter_parsing_state in parameter_parsing_states:
        content_parser_parameter = parameter_parsing_state.content_parser_parameter
        if content_parser_parameter.is_kwargs:
            return parameter_parsing_state
        
        if content_parser_parameter.display_name == keyword:
            return parameter_parsing_state
    
    return None


def get_rest_parameter_state(parameter_parsing_states):
    """
    Gets the rest parameter from the given content if there is any.
    
    Parameters
    ----------
    parameter_parsing_states `list` of ``ParameterParsingStateBase`` instances
        The created parameter parser state instances.

    Returns
    -------
    parameter_parsing_state : ``ParameterParsingState`` or `None`
    """
    for parameter_parsing_state in parameter_parsing_states:
        if parameter_parsing_state.content_parser_parameter.is_rest:
            return parameter_parsing_state


def create_parameter_parsing_states(command_content_parser):
    """
    Creates parser states for the given content parser parameters.
    
    Parameters
    ----------
    command_content_parser : ``CommandContentParser``
        The parameters of a ``CommandContentParser``.
    
    Returns
    -------
    parameter_parsing_states : `list` of ``ParameterParsingStateBase`` instances
        The created parameter parser state instances.
    """
    parameter_parsing_states = []
    for content_parser_parameters in command_content_parser._parameters:
        if content_parser_parameters.is_kwargs:
            parameter_parser_state_type = KwargsParameterParsingState
        elif content_parser_parameters.is_rest:
            parameter_parser_state_type = RestParameterParsingState
        else:
            parameter_parser_state_type = GenericParameterParsingState
        
        parameter_parser_state = parameter_parser_state_type(content_parser_parameters)
        parameter_parsing_states.append(parameter_parser_state)
    
    return parameter_parsing_states


def get_is_all_parameter_parser_state_satisfied(parameter_parsing_states):
    """
    Returns whether all the parser states are satisfied.
    
    Parameters
    ----------
    parameter_parsing_states : `list` of ``ParameterParsingStateBase`` instances
        The used parameter parsing states.
    
    Returns
    -------
    is_all_parameter_parser_state_satisfied : `bool`
    """
    for parameter_parser_state in parameter_parsing_states:
        if not parameter_parser_state.is_satisfied():
            return False
    
    return True


class ParameterParsingStateBase:
    """
    Base class for parsing states used meanwhile parsing parameters of a command.
    
    Attributes
    ----------
    content_parser_parameter : ``ContentParserParameter``
        The respective parameter.
    """
    __slots__ = ('content_parser_parameter', )
    def __new__(cls, content_parser_parameter):
        """
        Creates a new ``ParameterParsingState`` from the given ``content_parser_parameter`` instance.
        
        Parameters
        ----------
        content_parser_parameter : ``ContentParserParameter``
            The respective parameter.
        """
        self = object.__new__(cls)
        self.content_parser_parameter = content_parser_parameter
        return self
    
    
    def add_parsed_value(self, parsed_value, keyword):
        """
        Adds a value to the ``ParameterParsingState``.
        
        Parameters
        ----------
        parsed_value : `Any`
            The parsed value.
        keyword : `None` or `str`
            The keyword used to reference the value.
        """
        pass
    
    
    def get_parser_value(self, args, kwargs):
        """
        Gets the parser's value.
        
        Parameters
        ----------
        args : `list`
            Parameters to pass to a command's function.
        kwargs : `dict`
            Keyword parameter to pass to a command's function.

        Raises
        ------
        CommandParameterParsingError
            Parsing was unsuccessful.
        """
        pass

    
    def is_satisfied(self):
        """
        Returns whether the respective parameter is satisfied.
        
        Returns
        -------
        is_satisfied : `bool`
        """
        return True
    
    
    def __repr__(self):
        """Returns the parsing state's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' content_parser_parameter=',
            repr(self.content_parser_parameter),
            '>',
        ]
        
        return ''.join(repr_parts)


class GenericParameterParsingState(ParameterParsingStateBase):
    """
    Generic parsing state used meanwhile parsing parameters of a command.
    
    Attributes
    ----------
    content_parser_parameter : ``ContentParserParameter``
        The respective parameter.
    satisfied : `bool`
        Whether the parser is satisfied.
    parsed_values : `None` or `list` of `Any`
        The parsed values for the respective ``ContentParserParameter``.
    """
    __slots__ = ('satisfied', 'parsed_values', )
    
    @copy_docs(ParameterParsingStateBase.__new__)
    def __new__(cls, content_parser_parameter):
        self = object.__new__(cls)
        self.content_parser_parameter = content_parser_parameter
        self.parsed_values = None
        self.satisfied = False
        return self
    
    
    @copy_docs(ParameterParsingStateBase.add_parsed_value)
    def add_parsed_value(self, parsed_value, keyword):
        parsed_values = self.parsed_values
        if parsed_values is None:
            self.parsed_values = parsed_values = []
        
        parsed_values.append(parsed_value)
        
        if (parsed_value is not None):
            content_parser_parameter = self.content_parser_parameter
            if content_parser_parameter.is_positional or content_parser_parameter.is_keyword:
                self.satisfied = True
    
    
    @copy_docs(ParameterParsingStateBase.get_parser_value)
    def get_parser_value(self, args, kwargs):
        content_parser_parameter = self.content_parser_parameter
        if content_parser_parameter.is_positional:
            parsed_value = get_first_non_none_parsed_value(self.parsed_values)
            if parsed_value is None:
                if (not content_parser_parameter.has_default):
                    raise CommandParameterParsingError(content_parser_parameter)
                
                parsed_value = content_parser_parameter.default
                
            args.append(parsed_value)
            return
        
        if content_parser_parameter.is_keyword:
            parsed_value = get_first_non_none_parsed_value(self.parsed_values)
            if parsed_value is None:
                if (not content_parser_parameter.has_default):
                    raise CommandParameterParsingError(content_parser_parameter)
                
                parsed_value = content_parser_parameter.default
            
            kwargs[content_parser_parameter.name] = parsed_value
            return
        
        if content_parser_parameter.is_args:
            parsed_values = get_all_non_none_parsed_value(self.parsed_values)
            args.extend(parsed_values)
    
    
    @copy_docs(ParameterParsingStateBase.is_satisfied)
    def is_satisfied(self):
        return self.satisfied
    
    
    @copy_docs(ParameterParsingStateBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' content_parser_parameter=',
            repr(self.content_parser_parameter),
            ', parsed_values=',
            repr(self.parsed_values),
            '>',
        ]
        
        return ''.join(repr_parts)


class KwargsParameterParsingState(ParameterParsingStateBase):
    """
    `**kwargs`` specific parsing state used meanwhile parsing parameters of a command.
    
    Attributes
    ----------
    content_parser_parameter : ``ContentParserParameter``
        The respective parameter.
    parsed_items : `None` or `list` of `tuple` (`str`, `Any`)
        The parsed values for the respective ``ContentParserParameter``.
    """
    __slots__ = ('parsed_items', )
    
    @copy_docs(ParameterParsingStateBase.__new__)
    def __new__(cls, content_parser_parameter):
        self = object.__new__(cls)
        self.content_parser_parameter = content_parser_parameter
        self.parsed_items = None
        return self
    
    
    @copy_docs(ParameterParsingStateBase.add_parsed_value)
    def add_parsed_value(self, parsed_value, keyword):
        parsed_items = self.parsed_items
        if parsed_items is None:
            self.parsed_items = parsed_items = []
        
        parsed_items.append((keyword, parsed_value))
    
    
    @copy_docs(ParameterParsingStateBase.get_parser_value)
    def get_parser_value(self, args, kwargs):
        parsed_items = self.parsed_items
        if (parsed_items is not None):
            for key, value in parsed_items:
                if (value is not None):
                    kwargs.setdefault(key, value)
    
    
    @copy_docs(ParameterParsingStateBase.is_satisfied)
    def is_satisfied(self):
        return False
    
    
    @copy_docs(ParameterParsingStateBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' content_parser_parameter=',
            repr(self.content_parser_parameter),
            ', parsed_items=',
            repr(self.parsed_items),
            '>',
        ]
        
        return ''.join(repr_parts)


class RestParameterParsingState(ParameterParsingStateBase):
    """
    Parameter parsing state which consumes all the unused content. Used parsing state used meanwhile parsing
    parameters of a command.
    
    Attributes
    ----------
    content_parser_parameter : ``ContentParserParameter``
        The respective parameter.
    value : `None` or `str`
        The set rest value.
    """
    __slots__ = ('value', )
    
    @copy_docs(ParameterParsingStateBase.__new__)
    def __new__(cls, content_parser_parameter):
        self = object.__new__(cls)
        self.content_parser_parameter = content_parser_parameter
        self.value = None
        return self
    
    
    @copy_docs(ParameterParsingStateBase.add_parsed_value)
    def add_parsed_value(self, parsed_value, keyword):
        if parsed_value:
            self.value = parsed_value
    
    
    @copy_docs(ParameterParsingStateBase.get_parser_value)
    def get_parser_value(self, args, kwargs):
        parsed_value = self.value
        if parsed_value is None:
            content_parser_parameter = self.content_parser_parameter
            if content_parser_parameter.has_default:
                parsed_value = content_parser_parameter.default
            else:
                parsed_value = ''
        
        args.append(parsed_value)
    
    
    @copy_docs(ParameterParsingStateBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' content_parser_parameter=',
            repr(self.content_parser_parameter),
            ', value=',
            repr(self.value),
            '>',
        ]
        
        return ''.join(repr_parts)

def get_first_non_none_parsed_value(parsed_values):
    """
    Gets the first non-`None` parsed value.
    
    Parameters
    ----------
    parsed_values : `None` or `list` of `Any`
        The parsed value.

    Returns
    -------
    value : `None` or `Any`
        The parsed value if any.
    """
    if (parsed_values is not None):
        for parsed_value in parsed_values:
            if (parsed_value is not None):
                return parsed_value
    
    return None


def get_all_non_none_parsed_value(parsed_values):
    """
    Gets all the non-`None` parsed values.
    
    Parameters
    ----------
    parsed_values : `None` or `list` of `Any`
        The parsed value.

    Returns
    -------
    values : `None` or `Any`
        The parsed value if any.
    """
    values = []
    if (parsed_values is not None):
        for parsed_value in parsed_values:
            if (parsed_value is not None):
                values.append(parsed_value)
    
    return values


def content_parser_parameter_postprocessor_try_find_context(command_context_parser):
    """
    Tries to find context variable inside of a command context parser's parameters.
    
    Parameters
    ----------
    command_context_parser : ``CommandContentParser``
        The respective command content parser.
    """
    parameters = command_context_parser._parameters
    for parameter in parameters:
        detail = parameter.detail
        if (detail is not None):
            if detail.converter_setting is CONVERTER_SELF_CONTEXT:
                return
    
    for parameter in parameters:
        detail = parameter.detail
        if (detail is not None):
            if detail.converter_setting is CONVERTER_NONE and parameter.name in ('ctx', 'context', 'command_context'):
                parameter.set_converter_setting(CONVERTER_SELF_CONTEXT)
                return


def content_parser_parameter_postprocessor_try_find_message_and_client(command_context_parser):
    """
    Tries to find client and the message variables inside of a command context parser's parameters.
    
    Parameters
    ----------
    command_context_parser : ``CommandContentParser``
        The respective command content parser.
    """
    parameters = command_context_parser._parameters
    if len(parameters) < 2:
        return
    
    parameter_1, parameter_2 = parameters[:2]
    if parameter_1.name not in ('client', 'clnt', 'c'):
        return
    
    if parameter_2.name not in ('message', 'msg', 'm'):
        return
    
    detail = parameter_1.detail
    if detail is None:
        return
    
    converter_setting = detail.converter_setting
    if (converter_setting is not CONVERTER_NONE) and (converter_setting is not CONVERTER_CLIENT):
        return
    
    detail = parameter_2.detail
    if detail is None:
        return
    
    converter_setting = detail.converter_setting
    if (converter_setting is not CONVERTER_NONE) and (converter_setting is not CONVERTER_MESSAGE):
        return
    
    parameter_1.set_converter_setting(CONVERTER_SELF_CLIENT)
    parameter_2.set_converter_setting(CONVERTER_SELF_MESSAGE)


def content_parser_parameter_postprocessor_try_find_rest_parser(command_context_parser):
    """
    Tries to find rest parser
    
    Parameters
    ----------
    command_context_parser : ``CommandContentParser``
        The respective command content parser.
    """
    parameters = command_context_parser._parameters
    if not parameters:
        return
    
    for parameter in parameters:
        if parameter.is_args or parameter.is_kwargs or parameter.is_keyword:
            return
    
    parameter = parameters[-1]
    detail = parameter.detail
    if (detail is None):
        return
    
    if (detail.converter_setting is not CONVERTER_NONE):
        return
    
    parameter.is_rest = True


def convert_parser_parameter_postprocessor_try_find_string(command_context_parser):
    """
    Sets the unset parameters to use string converters.
    
    Parameters
    ----------
    command_context_parser : ``CommandContentParser``
        The respective command content parser.
    """
    parameters = command_context_parser._parameters
    for parameter in parameters:
        detail = parameter.detail
        if (detail is None):
            continue
        
        if (detail.converter_setting is CONVERTER_NONE) and (not parameter.is_rest):
            parameter.set_converter_setting(CONVERTER_STR)


CONTENT_PARSER_PARAMETER_POSTPROCESSORS = [
    content_parser_parameter_postprocessor_try_find_context,
    content_parser_parameter_postprocessor_try_find_message_and_client,
    content_parser_parameter_postprocessor_try_find_rest_parser,
    convert_parser_parameter_postprocessor_try_find_string,
]



from .client_extension import *
from .exceptions import *
from .extension import *
from .extension_loader import *
from .snapshot import *
from .utils import *

__all__ = (
    *client_extension.__all__,
    *exceptions.__all__,
    *extension.__all__,
    *extension_loader.__all__,
    *snapshot.__all__,
    *utils.__all__,
)

from .. import register_library_extension
register_library_extension('HuyaneMatsu.extension_loader')
del register_library_extension



__all__ = ('ICON_TYPE_ANIMATED', 'ICON_TYPE_NONE', 'ICON_TYPE_STATIC', 'Icon', 'IconType', 'IconSlot',
    'PreinstancedBase', )

import sys

from ...backend.utils import DOCS_ENABLED, doc_property
from .preinstanced import PreinstancedBase, Preinstance as P

class IconType(PreinstancedBase):
    """
    Represents a Discord icon's type.
    
    Attributes
    ----------
    name : `str`
        The name of the icon type.
    value : `int`
        The identifier value the icon type.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``IconType``) items
        Stores the predefined ``IconType`` instances. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The icon types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the icon types.
    
    Every predefined icon type can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | name          | value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | static                | static        | 1     |
    +-----------------------+---------------+-------+
    | animated              | animated      | 2     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    def __bool__(self):
        """Returns whether the icon's type is set."""
        if self.value:
            boolean = True
        else:
            boolean = False
        
        return boolean
    
    none = P(0, 'none')
    static = P(1, 'static')
    animated = P(2, 'animated')


ICON_TYPE_NONE = IconType.none
ICON_TYPE_STATIC = IconType.static
ICON_TYPE_ANIMATED = IconType.animated

class Icon:
    """
    Represents a Discord Icon.
    
    Attributes
    ----------
    hash : `int`
        The icon's hash value.
    type : ``IconType``
        The icon's type.
    """
    __slots__ = ('type', 'hash',)
    
    def __init__(self, icon_type, icon_hash):
        """
        Creates a new ``Icon`` object with the given attributes.
        
        Parameters
        ----------
        icon_type : ``IconType``
            The icon's type.
        icon_hash : `int`
            The icon's hash value.
        """
        self.type = icon_type
        self.hash = icon_hash
        
    @property
    def as_base16_hash(self):
        """
        Returns the discord side representation of the icon.
        
        Returns
        -------
        icon : `None` or `str`
        """
        icon_type = self.type
        if icon_type is ICON_TYPE_NONE:
            icon = None
        else:
            icon = self.hash.__format__('0>32x')
            if icon_type is ICON_TYPE_ANIMATED:
                icon = 'a_'+icon
        
        return icon
    
    hash_info_width = sys.hash_info.width
    if hash_info_width == 32:
        def __hash__(self):
            """Returns the icon's hash."""
            icon_type = self.type
            if icon_type is ICON_TYPE_NONE:
                hash_value = 0
            else:
                icon_hash = self.hash
                hash_value = (icon_hash>>96)^ \
                             ((icon_hash>>64)&((1<<32)-1))^ \
                             ((icon_hash>>32)&((1<<32)-1))^ \
                             (icon_hash&((1<<32)-1))
                
                if icon_type is ICON_TYPE_ANIMATED:
                    hash_value ^= ((1<<32)-1)
            
            return hash_value
    
    elif hash_info_width == 64:
        def __hash__(self):
            """Returns the icon's hash."""
            icon_type = self.type
            if icon_type is ICON_TYPE_NONE:
                hash_value = 0
            else:
                icon_hash = self.hash
                hash_value = (icon_hash>>64)^(icon_hash&((1<<64)-1))
                if icon_type is ICON_TYPE_ANIMATED:
                    hash_value ^= ((1<<64)-1)
            return hash_value
    
    else:
        def __hash__(self):
            """Returns the icon's hash."""
            icon_type = self.type
            if icon_type is ICON_TYPE_NONE:
                hash_value = 0
            else:
                hash_value = self.hash
                if icon_type is ICON_TYPE_ANIMATED:
                    hash_value ^= ((1<<128)-1)
            
            return hash_value
    
    del hash_info_width
    
    def __eq__(self, other):
        """Returns whether the two icons are equal."""
        if (type(self) is not type(other)):
            return NotImplemented
        
        icon_type = self.type
        if (icon_type is not other.type):
            return False
        
        if icon_type is ICON_TYPE_NONE:
            return True
        
        if self.hash == other.hash:
            return True
        
        return False
    
    def __repr__(self):
        """Returns the representation of the icon."""
        return f'{self.__class__.__name__}(type=ICON_TYPE_{self.type.name.upper()}, hash={self.hash})'
    
    def __iter__(self):
        """
        Unpacks the icon.
        
        This method is a generator.
        """
        yield self.type
        yield self.hash
    
    def __len__(self):
        """Length hinter (for unpacking if needed)."""
        return 2
    
    def __bool__(self):
        return (self.type is not ICON_TYPE_NONE)
    
    @classmethod
    def from_base16_hash(cls, icon):
        """
        Converts a discord icon hash value to an ``Icon`` object.
        
        Parameters
        ----------
        icon : `None` or `str`
        
        Returns
        -------
        self : ``Icon``
        """
        if icon is None:
            icon_type = ICON_TYPE_NONE
            icon_hash = 0
        else:
            if icon.startswith('a_'):
                icon = icon[2:]
                icon_type = ICON_TYPE_ANIMATED
            else:
                icon_type = ICON_TYPE_STATIC
            icon_hash = int(icon, 16)
        
        self = object.__new__(cls)
        self.type = icon_type
        self.hash = icon_hash
        return self


class IconSlot:
    if DOCS_ENABLED:
        __class_doc__ = (
    """
    Internal icon slotter to represent an icon of a discord entity.
    
    Attributes
    ----------
    internal_name : `str`
        The internal name of the icon.
    discord_side_name : `str`
        The discord side name of the icon.
    added_instance_attributes : `tuple` of `str`
        The added instance attribute's name by the icon slot.
    added_class_attributes : `list` of `tuple` (`str`, `Any`)
        The added class attributes by the icon slot.
    
    Class Attributes
    ----------------
    _compile_globals : `dict` of (`str`, `Any`)
        Compile time globals for the generated functions.
    """)
        
        __instance_doc__ = (
    """
    Returns the respective icon.
    
    Returns
    -------
    icon : ``Icon``
    """)
        
        __doc__ = doc_property()
    
    __slots__ = ('internal_name', 'discord_side_name', 'added_instance_attributes', 'added_class_attributes')
    
    _compile_globals = {
        'ICON_TYPE_NONE': ICON_TYPE_NONE     ,
        'ICON_TYPE_STATIC': ICON_TYPE_STATIC   ,
        'ICON_TYPE_ANIMATED': ICON_TYPE_ANIMATED ,
        'Icon': Icon               ,
    }
    
    def __new__(cls, internal_name, discord_side_name, url_property, url_as_method, add_updater=True):
        """
        Creates an ``IconSlot`` with the given parameters.
        
        Parameters
        ----------
        internal_name : `str`
            The internal name of the icon.
        discord_side_name : `str`
            The discord side name of the icon.
        url_property : `None` or `function`
            A function what will be used as a property when accessing the icon' url.
        url_as_method : `None` or `function`
            A function what will be used a method when creating a formatted icon url.
        add_updater : `bool`, Optional
            Whether the icon slot should add updater methods to the class. Defaults to `True`.
        
        Returns
        -------
        self : ``IconSlot``
        """
        added_instance_attribute_name_hash = internal_name+'_hash'
        added_internal_attribute_name_type = internal_name+'_type'
        
        added_class_attributes = []
        if (url_property is not None):
            added_class_attributes.append((f'{internal_name}_url', property(url_property)))
        
        if (url_as_method is not None):
            added_class_attributes.append((f'{internal_name}_url_as', url_as_method))
        
        locals_ = {}
        func_name = f'_set_{internal_name}'
        exec(compile((
            f'def {func_name}(self, data):\n'
            f'    icon = data.get({discord_side_name!r}, None)\n'
            f''
            f'    if icon is None:\n'
            f'        icon_type = ICON_TYPE_NONE\n'
            f'        icon_hash = 0\n'
            f'    else:\n'
            f'        if icon.startswith(\'a_\'):\n'
            f'            icon = icon[2:]\n'
            f'            icon_type = ICON_TYPE_ANIMATED\n'
            f'        else:\n'
            f'            icon_type = ICON_TYPE_STATIC\n'
            f'        icon_hash = int(icon, 16)\n'
            f''
            f'    self.{added_internal_attribute_name_type} = icon_type\n'
            f'    self.{added_instance_attribute_name_hash} = icon_hash\n'
        ), f'<{cls.__name__}>', 'exec', optimize=2), cls._compile_globals, locals_)
        
        added_class_attributes.append((func_name, locals_[func_name]),)
        
        if add_updater:
            locals_ = {}
            func_name = f'_update_{internal_name}'
            exec(compile((
                f'def {func_name}(self, data, old_attributes):\n'
                f'    icon = data.get({discord_side_name!r}, None)\n'
                f''
                f'    if icon is None:\n'
                f'        icon_type = ICON_TYPE_NONE\n'
                f'        icon_hash = 0\n'
                f'    else:\n'
                f'        if icon.startswith(\'a_\'):\n'
                f'            icon = icon[2:]\n'
                f'            icon_type = ICON_TYPE_ANIMATED\n'
                f'        else:\n'
                f'            icon_type = ICON_TYPE_STATIC\n'
                f'        icon_hash = int(icon, 16)\n'
                f''
                f'    self_icon_type = self.{added_internal_attribute_name_type}\n'
                f'    self_icon_hash = self.{added_instance_attribute_name_hash}\n'
                f'    if (self_icon_type is not icon_type) or (self_icon_hash != icon_hash):\n'
                f'        old_attributes[{internal_name!r}] = Icon(self_icon_type, self_icon_hash)\n'
                f'        self.{added_internal_attribute_name_type} = icon_type\n'
                f'        self.{added_instance_attribute_name_hash} = icon_hash\n'
            ), f'<{cls.__name__}>', 'exec', optimize=2), cls._compile_globals, locals_)
            
            added_class_attributes.append((func_name, locals_[func_name]),)
        
        self = object.__new__(cls)
        self.internal_name = internal_name
        self.discord_side_name = discord_side_name
        self.added_instance_attributes = (added_internal_attribute_name_type, added_instance_attribute_name_hash)
        self.added_class_attributes = added_class_attributes
        return self
    
    def __set_slot__(self, attribute_name, class_attributes, class_slots):
        """Applies the changes of the icon slot on the class's attributes."""
        
        # Extend the slots of the class
        class_slots.update(self.added_instance_attributes)
        
        # Add the extra class attributes to the class
        for name, value in self.added_class_attributes:
            class_attributes[name] = value
    
    def __get__(self, obj, type_):
        """Returns self if called from class, meanwhile an ``Icon`` if called from an object."""
        if obj is None:
            return self
        
        icon_type_name, icon_hash_name = self.added_instance_attributes
        icon_type = getattr(obj, icon_type_name)
        icon_hash = getattr(obj, icon_hash_name)
        return Icon(icon_type, icon_hash)
    
    def __set__(self, obj, value):
        """Can't set attribute."""
        raise AttributeError('can\'t set attribute')
    
    def __delete__(self, obj):
        """Can't delete attribute."""
        raise AttributeError('can\'t delete attribute')
    
    def preconvert(self, kwargs, processable):
        """
        Used at preconverters to parse out from the passed kwargs the icon of the entity.
        
        Parameters
        ----------
        kwargs : `dict` of (`str`, `Any`) items
            Keyword parameters passed to the respective preconverter.
        processable : `list` of `tuple` (`str`, `Any`)
            A list of instance attributes which will be set when all the passed kwargs are validated.
        
        Raises
        ------
        TypeError
            If any of expected value's type is invalid.
        ValueError
            If any of the expected value's type is valid, but it's value is not.
        """
        icon_type_name, icon_hash_name = self.added_instance_attributes
        try:
            icon = kwargs.pop(self.internal_name)
        except KeyError:
            try:
                icon_hash = kwargs.pop(icon_hash_name)
            except KeyError:
                return
            
            if type(icon_hash) is int:
                pass
            elif isinstance(icon_hash, int):
                icon_hash = int(icon_hash)
            else:
                raise TypeError(f'`{icon_hash_name}` can be passed as `int` instance, got '
                    f'{icon_hash.__class__.__name__}.')
            
            if icon_hash < 0 or icon_hash > ((1<<128)-1):
                raise ValueError(f'`{icon_hash_name}` cannot be negative or longer than 128 bits, got {icon_hash}.')
            
            try:
                icon_type = kwargs.pop(icon_type_name)
            except KeyError:
                if icon_hash == 0:
                    icon_type = ICON_TYPE_NONE
                else:
                    icon_type = ICON_TYPE_STATIC
            else:
                if (type(icon_type) is not IconType):
                    raise TypeError(f'`{icon_type_name}` can be passed as `{IconType.__name__}` instance, got '
                        f'{icon_type.__class__.__name__}.')
                
                if (icon_type is ICON_TYPE_NONE) and icon_hash:
                    raise ValueError(f'If `{icon_type_name}` is passed as `ICON_TYPE_NONE`, then `{icon_hash_name}` '
                        f'can be passed only as `0`, meanwhile got `{icon_hash}`.')
            
        else:
            if icon is None:
                icon_type = ICON_TYPE_NONE
                icon_hash = 0
            elif type(icon) is Icon:
                icon_type = icon.type
                icon_hash = icon.hash
            elif isinstance(icon, str):
                if icon.startswith('a_'):
                    icon = icon[2:]
                    icon_type = ICON_TYPE_ANIMATED
                else:
                    icon_type = ICON_TYPE_STATIC
                icon_hash = int(icon, 16)
            else:
                raise TypeError(f'`{self.internal_name!r}` can be passed as `None`, `{Icon.__name__}` or as `str` '
                    f'instance, got {icon.__class__.__name__}.')
        
        processable[icon_type_name] = icon_type
        processable[icon_hash_name] = icon_hash



__all__ = ('VoiceState', )

from datetime import datetime

from ...backend.export import include

from ..utils import timestamp_to_datetime
from ..core import CHANNELS, GUILDS

from .utils import create_partial_user_from_id

create_partial_role_from_id = include('create_partial_role_from_id')

class VoiceState:
    """
    Represents a user at a ``ChannelVoice``.
    
    Attributes
    ----------
    _cache_user : ``ClientUserBase``
        The voice state's respective user.
    channel_id : `int`
        The channel's identifier to where the user is connected to.
    deaf : `bool`
        Whether the user is deafen.
    guild_id : `int`
        The guild's identifier where the user is connected to.
    is_speaker : `bool`
        Whether the user is suppressed inside of the voice channel.
        
        If the channel is a ``ChannelVoice``, it is always `False`, meanwhile it ``ChannelStage`` it can vary.
    
    mute : `bool`
        Whether the user is muted.
    requested_to_speak_at : `None` or `datetime`
        When the user requested to speak.
        
        Only applicable if the user is connected to a ``ChannelStage`` instance.
    
    self_deaf : `bool`
        Whether the user muted everyone else.
    self_mute : `bool`
        Whether the user muted itself.
    self_stream : `bool`
        Whether the user screen shares with the go live option.
    self_video : `bool`
        Whether the user sends video from a camera source.
    session_id : `str``
        The user's voice session id.
    user_id : `int`
        The voice state's respective user's identifier.
    """
    __slots__ = ('_cache_user', 'channel_id', 'deaf', 'guild_id', 'is_speaker', 'mute', 'requested_to_speak_at',
        'self_deaf', 'self_mute', 'self_stream', 'self_video', 'session_id', 'user_id')
    
    def __new__(cls, data, guild_id):
        """
        Creates a ``VoiceState`` object from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Voice state data received from Discord.
        guild_id : `int`
            The voice state's guild's identifier.
        """
        channel_id = data.get('channel_id', None)
        if channel_id is None:
            return
        
        channel_id = int(channel_id)
        
        user_id = int(data['user_id'])
        
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            guild = None
        else:
            try:
                return guild.voice_states[user_id]
            except KeyError:
                pass
        
        self = object.__new__(cls)
        self.channel_id = channel_id
        self.guild_id = guild_id
        self.user_id = user_id
        self.session_id = data['session_id']
        self.mute = data['mute']
        self.deaf = data['deaf']
        self.self_deaf = data['self_deaf']
        self.self_mute = data['self_mute']
        self.self_stream = data.get('self_stream', False)
        self.self_video = data['self_video']
        
        requested_to_speak_at = data.get('request_to_speak_timestamp', None)
        if (requested_to_speak_at is not None):
            requested_to_speak_at = timestamp_to_datetime(requested_to_speak_at)
        
        self.is_speaker = not data.get('suppress', False)
        
        self.requested_to_speak_at = requested_to_speak_at
        
        if (guild is not None):
            guild.voice_states[user_id] = self
        
        return self
    
    
    @property
    def user(self):
        """
        Returns the voice state's user.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        user = self._cache_user
        if (user is None):
            user = create_partial_user_from_id(self.user_id)
            self._cache_user = user
        
        return user
    
    def _set_cache_user(self, user):
        """
        Sets the cached user of the voice state.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The respective user of the voice state.
        """
        self._cache_user = user
    
    @property
    def channel(self):
        """
        Returns the voice state's channel, where the user is connected to.
        
        Returns
        -------
        channel : `None` or ``ChannelVoiceBase``
        """
        channel_id = self.channel_id
        if channel_id:
            return CHANNELS[channel_id]
    
    
    @property
    def guild(self):
        """
        Returns the voice state's respective guild.
        
        Returns
        -------
        guild : `None` or ``Guild``
        """
        channel = self.channel
        if (channel is not None):
            return channel.guild
    
    
    def _update_channel(self, data):
        """
        Updates the voice state's channel.
        
        Returns
        -------
        old_channel_id : `int`
            The voice state's old channel's identifier.
        new_channel_id : `int`
            The voice state's new channel's identifier.
        """
        channel_id = data.get('channel_id', None)
        if channel_id is None:
            channel_id = 0
            
            try:
                guild = GUILDS[self.guild_id]
            except KeyError:
                pass
            else:
                try:
                    del guild.voice_states[self.user_id]
                except KeyError:
                    pass
        else:
            channel_id = int(channel_id)
        
        old_channel_id = self.channel_id
        self.channel_id = channel_id
        
        return old_channel_id, channel_id
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the voice state and returns it's overwritten attributes as a `dict` with a `attribute-name` -
        `old-value` relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Voice state data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dictionary is optional.
        
        Returned Data Structure
        -----------------------
        +-----------------------+-----------------------+
        | Keys                  | Values                |
        +=======================+=======================+
        | deaf                  | `str`                 |
        +-----------------------+-----------------------+
        | is_speaker            | `bool`                |
        +-----------------------+-----------------------+
        | mute                  | `bool`                |
        +-----------------------+-----------------------+
        | requested_to_speak_at | `None` or `datetime`  |
        +-----------------------+-----------------------+
        | self_deaf             | `bool`                |
        +-----------------------+-----------------------+
        | self_mute             | `bool`                |
        +-----------------------+-----------------------+
        | self_stream           | `bool`                |
        +-----------------------+-----------------------+
        | self_video            | `bool`                |
        +-----------------------+-----------------------+
        """
        old_attributes = {}
        
        deaf = data['deaf']
        if self.deaf != deaf:
            old_attributes['deaf'] = self.deaf
            self.deaf = deaf
        
        mute = data['mute']
        if self.mute != mute:
            old_attributes['mute'] = self.mute
            self.mute = mute
        
        self_deaf = data['self_deaf']
        if self.self_deaf != self_deaf:
            old_attributes['self_deaf'] = self.self_deaf
            self.self_deaf = self_deaf
        
        self_video = data['self_video']
        if self.self_video != self_video:
            old_attributes['self_video'] = self.self_video
            self.self_video = self_video
        
        self_stream = data.get('self_stream', False)
        if self.self_stream != self_stream:
            old_attributes['self_stream'] = self.self_stream
            self.self_stream = self_stream
        
        self_mute = data['self_mute']
        if self.self_mute != self_mute:
            old_attributes['self_mute'] = self.self_mute
            self.self_mute = self_mute
        
        requested_to_speak_at = data.get('request_to_speak_timestamp', None)
        if (requested_to_speak_at is not None):
            requested_to_speak_at = timestamp_to_datetime(requested_to_speak_at)
        
        if self.requested_to_speak_at != requested_to_speak_at:
            old_attributes['requested_to_speak_at'] = self.requested_to_speak_at
            self.requested_to_speak_at = requested_to_speak_at
        
        is_speaker = not data.get('suppress', False)
        if self.is_speaker != is_speaker:
            old_attributes['is_speaker'] = self.is_speaker
            self.is_speaker = is_speaker
        
        return old_attributes
    
    
    def _update_attributes(self, data):
        """
        Updates the voice state with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Voice state data received from Discord.
        """
        self.deaf = data['deaf']
        self.mute = data['mute']
        self.self_deaf = data['self_deaf']
        self.self_mute = data['self_mute']
        self.self_stream = data.get('self_stream', False)
        self.self_video = data['self_video']
        
        requested_to_speak_at = data.get('request_to_speak_timestamp', None)
        if (requested_to_speak_at is not None):
            requested_to_speak_at = timestamp_to_datetime(requested_to_speak_at)
        
        self.requested_to_speak_at = requested_to_speak_at
        
        self.is_speaker = not data.get('suppress', False)
    
    def __repr__(self):
        """Returns the voice state's representation."""
        return (
            f'<{self.__class__.__name__} user_id={self.user_id}, guild_id={self.guild_id}, '
            f'channel_id={self.channel_id}>'
        )



__all__ = ()

from .color import Color

def preconvert_snowflake(snowflake, name):
    """
    Converts the given `snowflake` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    snowflake : `str` or `int`
        The snowflake to convert.
    name : `str`
        The name of the snowflake.
    
    Returns
    -------
    snowflake : `int`
    
    Raises
    ------
    TypeError
        - If `snowflake` was not passed neither as `int` or as `str` instance.
    ValueError
        - If `snowflake` was passed as `str` and cannot be converted to `int`.
        - If the converted `snowflake` is negative or it's bit length is over 64.
    """
    snowflake_type = type(snowflake)
    if snowflake_type is int:
        pass
    if issubclass(snowflake_type, int):
        snowflake = int(snowflake)
    # JSON uint64 is str
    elif issubclass(snowflake_type, str):
        if 6 < len(snowflake) < 21 and snowflake.isdigit():
            snowflake = int(snowflake)
        else:
            raise ValueError(f'`{name}` can be passed as `int` or `str` instance, got `str` instance, but not a valid'
                f'snowflake (7-20 length, digit only), got {snowflake!r}.')
    else:
        raise TypeError(f'`{name}` can be passed as `int` or `str` instance, got {snowflake_type.__name__}.')
    
    if snowflake < 0 or snowflake > ((1<<64)-1):
        raise ValueError(f'`{name}` can be only uint64, got {snowflake!r}.')
    
    return snowflake


def preconvert_snowflake_array(snowflake_array, name):
    """
    Converts the given `snowflake_array` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    snowflake : `None` or (`tuple`, `list` or `set`) of (`str` or `int`)
        The snowflakes to convert.
    name : `str`
        The name of the snowflake array.
    Returns
    -------
    snowflake_array : `tuple` of `int`
        The returned value is always sorted.
    
    Raises
    ------
    TypeError
        - If `snowflake_array` is neither `None`, `list`, `tuple` nor `set`.
        - If `snowflake_array` contains a non `int` nor `str` instance.
    ValueError
        - If `snowflake_array`contains a `str`, what cannot be converted to `int`.
        - If a converted `snowflake` is negative or it's bit length is over 64.
    """
    if (snowflake_array is not None):
        if not isinstance(snowflake_array, (list, tuple, set)):
            raise TypeError(f'`{name}` can be either `list`, `tuple` or `set`, got '
                f'`{snowflake_array.__class__.__name__}`.')
        
        snowflake_array_processed = []
        
        for snowflake in snowflake_array:
            snowflake_type = type(snowflake)
            if snowflake_type is int:
                pass
            if issubclass(snowflake_type, int):
                snowflake = int(snowflake)
            # JSON uint64 is str
            elif issubclass(snowflake_type, str):
                if 6 < len(snowflake) < 21 and snowflake.isdigit():
                    snowflake = int(snowflake)
                else:
                    raise ValueError(f'`{name}`\'s elements can be passed as `int` or `str` instance, got `str` '
                        f'instance, but not a valid snowflake (7-20 length, digit only), got {snowflake!r}.')
            else:
                raise TypeError(f'`{name}`\'s elements can be passed as `int` or `str` instance, got '
                    f'{snowflake_type.__name__}.')
            
            if snowflake < 0 or snowflake > ((1<<64)-1):
                raise ValueError(f'`{name}`\'s elements can be only uint64, got {snowflake!r}.')
            
            snowflake_array_processed.append(snowflake)
        
        if snowflake_array_processed:
            snowflake_array_processed.sort()
            snowflake_array = tuple(snowflake_array_processed)
        else:
            snowflake_array = None
    
    return snowflake_array


def preconvert_discriminator(discriminator):
    """
    Converts the given `discriminator` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    discriminator : `str` or `int`
        The discriminator of an user to convert.
    
    Returns
    -------
    discriminator : `int`
    
    Raises
    ------
    TypeError
        If `discriminator` was not passed neither as `int` or as `str` instance.
    ValueError
        - If `discriminator` was passed as `str` and it is not numerical or it's length is over `4`.
        - If the `discriminator`'s value is less than `0` or is over than `9999`.
    """
    if (type(discriminator) is int):
        pass
    elif isinstance(discriminator, int):
        discriminator = int(discriminator)
    # Discord sends `discriminator` as `str`, so lets accept that as well.
    elif isinstance(discriminator, str):
        if 0<len(discriminator) < 5 and discriminator.isdigit():
            raise ValueError(f'`discriminator` was given as a `str` instance, but it is not numerical or it\'s length '
                 f'is over `4`, got {discriminator!r}')
        discriminator = int(discriminator)
    else:
        raise TypeError(f'`discriminator` can be passed as `int` or `str` instance, got '
            f'{discriminator.__class__.__name__}.')
    
    if discriminator < 0 or discriminator > 9999:
        raise ValueError(f'`discriminator` can be between 0 and 9999, got got {discriminator!r}.')
    
    return discriminator


def preconvert_color(color, name, nullable):
    """
    Converts the given `color` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    color : `int`
        The color to convert.
    name : `str`
        The name of the value.
    nullable : `bool`
        Whether the value is nullable.
    
    Returns
    -------
    color : `Color`
    
    Raises
    ------
    TypeError
        If `color` was not passed as `int` instance.
    ValueError
        If the `color`'s value is less than `0` or is over than `0xffffff`.
    """
    if not (nullable and (color is None)):
        if type(color) is Color:
            pass
        elif isinstance(color, int):
            color = Color(color)
        else:
            raise TypeError(f'`{name}` can be {"`None`, " if nullable else ""}`{Color.__name__}` or `int` instance, '
                f'got {color.__class__.__name__}.')
            
            if color < 0 or color > 0xffffff:
                raise ValueError(f'`{name}` can be between 0 and 0xffffff, got {color!r}.')
    
    return color


def preconvert_str(value, name, lower_limit, upper_limit):
    """
    Converts the given `value` to an acceptable string by the wrapper.
    
    Parameters
    ----------
    value : `str`
        The string to convert,
    name : `str`
        The name of the value.
    lower_limit : `int`
        The minimal length of the string.
    upper_limit : `int`
        The maximal length of the string.
    
    Returns
    -------
    value : `str`
    
    Raises
    ------
    TypeError
        If `value` was not passed as `str` instance.
    ValueError
        If the `value`'s length is less than the given `lower_limit` or is higher than the given than the given
        `upper_limit`.
    """
    if type(value) is str:
        pass
    elif isinstance(value, str):
        value = str(value)
    else:
        raise TypeError(f'`{name}` can be `str` instance, got {value.__class__.__name__}.')
    
    length = len(value)
    if (length != 0) and (length < lower_limit or length > upper_limit):
        raise ValueError(f'`{name}` can be between length {lower_limit} and {upper_limit}, got {length!r}; {value!r}.')
    
    return value


def preconvert_iterable_of_str(value, name, iterable_lower_limit, iterable_upper_limit, lower_limit, upper_limit):
    """
    Converts the given `value` to an acceptable iterable of string by the wrapper.
    
    Parameters
    ----------
    value : `str`
        The string to convert,
    name : `str`
        The name of the value.
    iterable_lower_limit : `int`
        The minimal length of `value`.
    iterable_upper_limit : `int`
        The maximal length of `value`.
    lower_limit : `int`
        The minimal length of the string.
    upper_limit : `int`
        The maximal length of the string.
    
    Returns
    -------
    converted_value : `set` of `str`
    
    Raises
    ------
    TypeError
        - If `value` was not passed as an `iterable`.
        - If `value` contains a non `str` element.
    ValueError
        - If the `value`'s length is less than the given `iterable_lower_limit` or is higher than the given than
            the given `iterable_upper_limit`.
        - If `value` contains a string, which's length is less than `lower_limit` or is more than `upper_limit`.
    """
    iterator = getattr(type(value), '__iter__', None)
    if iterator is None:
        raise TypeError(f'`{name}` can be `iterable`, got {value.__class__.__name__}.')
    
    converted_value = set()
    
    for value_element in iterator(value):
        if type(value_element) is str:
            pass
        elif isinstance(value_element, str):
            value_element = str(value_element)
        else:
            raise TypeError(f'`{name}` can contains `str` instances, got {value.__class__.__name__}.')
        
        length = len(value_element)
        if length == 0:
            continue
        
        if (length < lower_limit) or (length > upper_limit):
            raise ValueError(f'`{name}` can contains elements between length {lower_limit} and {upper_limit}, '
                f'got {length!r} {value_element!r}.')
        
        converted_value.add(value_element)
    
    length = len(converted_value)
    if (length != 0) and (length < iterable_lower_limit or length > iterable_upper_limit):
        raise ValueError(f'`{name} can be in between length {iterable_lower_limit} and {iterable_upper_limit}, '
            f'got {length!r}; {value!r}.')
    
    return converted_value


def preconvert_bool(value, name):
    """
    Converts the given `value` to an acceptable boolean by the wrapper.
    
    Parameters
    ----------
    value : `int`
        The value to convert.
    name : `str`
        The name of the value.
    
    Returns
    -------
    value : `str`
    
    Raises
    ------
    TypeError
        If `value` was not passed as `int` instance.
    ValueError
        If value was passed as an `int` instance, but not as `0` or `1` either.
    """
    if (type(value) is bool):
        pass
    elif isinstance(value, int):
        if (value not in (0, 1)):
            raise ValueError(f'`{name}` was given as `int` instance, but neither as `0` or `1`, got {value!r}.')
        value = bool(value)
    else:
        raise TypeError(f'`{name}` can be `bool` or `int` instance as `0` or `1`, got {value.__class__.__name__}.')
    
    return value


def preconvert_flag(flag, name, type_):
    """
    Converts the given `flag` to an acceptable flag by the wrapper.
    
    Parameters
    ----------
    flag : `int`
        The flag to convert.
    name : `str`
        The name of the flag.
    type_ : `int` subtype
        The type of the output flag.
    
    Returns
    -------
    flag : `type_`
    
    Raises
    ------
    TypeError
        If `flag` was not given as `int` instance.
    ValueError
        If `flag` was given as `int` instance, but it's value is less than `0` or it's bit length is over `64`.
    """
    flag_type = flag.__class__
    if (flag_type is type_):
        pass
    elif issubclass(flag_type, int):
        flag = type_(flag)
    else:
        raise TypeError(f'`{name}` can be passed as `{type_.__name__}` instance, got {flag_type.__name__}.')
    
    if flag < 0 or flag > ((1<<64)-1):
        raise ValueError(f'`{name}` can be only uint64, got {flag!r}.')
    
    return flag


def preconvert_preinstanced_type(value, name, type_):
    """
    Converts the given `value` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    type_ : ``PreinstancedBase`` instance
        The preinstanced type.
    
    Returns
    -------
    value : ``PreinstancedBase`` instance
    
    Raises
    ------
    TypeError
        If `value` was not given as `type_` instance, neither as `type_.value`'s type's instance.
    ValueError
        If there is no preinstanced object for the given `value`.
    """
    value_type = value.__class__
    if (value_type is not type_):
        value_expected_type = type_.VALUE_TYPE
        if value_type is value_expected_type:
            pass
        elif issubclass(value_type, value_expected_type):
            value = value_expected_type(value)
        else:
            raise TypeError(f'`{name}` can be passed as {type_.__name__} or as {value_expected_type.__name__} '
                f'instance, got {value_type.__name__}.')
        
        try:
            value = type_.INSTANCES[value]
        except LookupError:
            raise ValueError(f'There is no predefined `{name}` for the following value: {value!r}.') from None
    
    return value


def _pre_validate_int(value, name):
    """
    Converts the given `value` to `int`.

    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    
    Returns
    -------
    value : `int`
    
    Raises
    ------
    TypeError
        If `value` was not given as `int` instance.
    """
    if type(value) is int:
        pass
    elif isinstance(value, int):
        value = int(value)
    else:
        raise TypeError(f'`{name}` can be `int` instance, got {value.__class__.__name__}.')

    return value


def preconvert_int(value, name, lower_limit, upper_limit):
    """
    Converts the given `value` to an acceptable integer by the wrapper.
    
    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    lower_limit : `int`
        The minimal value of `value`.
    upper_limit : `int`
        The maximal value of `value`.
    
    Returns
    -------
    value : `int`
    
    Raises
    ------
    TypeError
        If `value` was not given as `int` instance.
    ValueError
        If `value` is less than `lower_limit`, or is higher than the `upper_limit`.
    """
    value = _pre_validate_int(value, name)
    
    if (value < lower_limit) or (value > upper_limit):
        raise ValueError(f'`{name}` can be between {lower_limit} and {upper_limit}, got {value!r}.')
    
    return value


def preconvert_int_options(value, name, options):
    """
    Converts the given `value` to an acceptable integer by the wrapper.
    
    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    options : `frozenset`
        The options, from which `value` should be one.
    
    Returns
    -------
    value : `int`
    
    Raises
    ------
    TypeError
        If `value` was not given as `int` instance.
    ValueError
        If `value` is less than `lower_limit`, or is higher than the `upper_limit`.
    """
    value = _pre_validate_int(value, name)
    
    if value not in options:
        raise ValueError(f'`{name}` can be any of: {", ".join(options)}, got {value!r}.')
    
    return value


def preconvert_float(value, name, lower_limit, upper_limit):
    """
    Converts the given `value` to an acceptable float by the wrapper.
    
    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    lower_limit : `float`
        The minimal value of `value`.
    upper_limit : `float`
        The maximal value of `value`.
    
    Returns
    -------
    value : `float`
    
    Raises
    ------
    TypeError
        If `value` was not given as `float` instance.
    ValueError
        If `value` is less than `lower_limit`, or is higher than the `upper_limit`.
    """
    if type(value) is float:
        pass
    elif isinstance(value, float):
        value = int(value)
    else:
        float_converter = getattr(type(value), '__float__', None)
        if float_converter is None:
            raise TypeError(f'`{name}` can be `float` instance, got {value.__class__.__name__}.')
        
        value = float_converter(value)
    
    if (value < lower_limit) or (value > upper_limit):
        raise ValueError(f'`{name}` can be between {lower_limit} and {upper_limit}, got {value!r}.')
    
    return value


def get_type_names(type_or_types):
    """
    Gets the given type(s)'s name closed within \` characters.
    
    Parameters
    ----------
    type_or_types : `type` or `tuple` of `type`
        The type(s) to get name of.
    
    Returns
    -------
    type_names : `str`
    """
    type_name_parts = []
    
    if isinstance(type_or_types, type):
        type_name_parts.append('`')
        type_name_parts.append(type_or_types.__name__)
        type_name_parts.append('`')
    else:
        length = len(type_or_types)
        if length:
            index = 0
            while True:
                type_ = type_or_types[index]
                index += 1
                
                type_name_parts.append('`')
                type_name_parts.append(type_.__name__)
                type_name_parts.append('`')
                
                if index == length:
                    break
                
                type_name_parts.append(', ')
                continue
    
    return ''.join(type_name_parts)



__all__ = ('CommandCheckError', 'CommandCooldownError', 'CommandParameterParsingError', 'CommandProcessingError', )

class CommandProcessingError(BaseException):
    """
    Exception raised when a command's processing failed.
    
    This is a superclass for various exceptions types raised around the table.
    """


class CommandParameterParsingError(CommandProcessingError):
    """
    Raised when a required parameter could not be parsed.
    
    Attributes
    ----------
    content_parser_parameter : ``ContentParserParameter``
        The unfulfilled parameter.
    """
    def __init__(self, content_parser_parameter):
        """
        Creates a new ``CommandParameterParsingError`` instance.
        
        Parameters
        ----------
        content_parser_parameter : ``ContentParserParameter``
            The unfulfilled parameter.
        """
        self.content_parser_parameter = content_parser_parameter
        CommandProcessingError.__init__(self, content_parser_parameter)


class CommandCheckError(CommandProcessingError):
    """
    Raised when a check fails.
    
    Attributes
    ----------
    check : ``CheckBase``
        The failed check.
    """
    def __init__(self, check):
        """
        Creates a new ``CommandCheckError`` instance.
        
        Parameters
        ----------
        check : ``ContentParserParameter``
            The failed check.
        """
        self.check = check
        CommandProcessingError.__init__(self, check)


class CommandCooldownError(CommandProcessingError):
    """
    Raised when the command is on cooldown.
    
    Attributes
    ----------
    expires_after : `float`
        After how much time the cooldown expires.
        
        If a prerequisite didn't pass, may be `-1.0`.
    
    cooldown_handler : ``CooldownHandler``
        The respective cooldown handler which is on cooldown.
    """
    def __init__(self, cooldown_handler, expires_after):
        """
        Creates a new ``CommandCheckError`` instance.
        
        Parameters
        ----------
        cooldown_handler : ``CooldownHandler``
            The respective cooldown handler which is on cooldown.
        expires_after : `float`
            After how much time the cooldown expires.
            
            If a prerequisite didn't pass, may be `-1.0`.
        """
        self.cooldown_handler = cooldown_handler
        self.expires_after = expires_after
        CommandProcessingError.__init__(self, cooldown_handler, expires_after)



__all__ = ('DiscoveryCategory', 'GuildDiscovery', )

from ..bases import DiscordEntity
from ..core import DISCOVERY_CATEGORIES
from ..utils import DISCORD_EPOCH_START, timestamp_to_datetime

class GuildDiscovery:
    """
    Represent a guild's Discovery settings.
    
    Attributes
    ----------
    application_actioned : `None` or `datetime`
        When the guild's application was accepted or rejected.
    application_requested : `None` or `datetime`
        When the guild applied to guild discovery. Only set if pending.
    emoji_discovery : `bool`
        Whether guild info is shown when the respective guild's emojis are clicked.
    guild : `Guild`
        The represented guild.
    keywords : `None` or (`set` of `str`)
        The set discovery search keywords for the guild.
    primary_category_id : `int`
        The `id` of the primary discovery category of the guild.
    sub_category_ids : `set` of `int`
        Guild Discovery sub-category id-s. Can be maximum 5.
    """
    __slots__ = ('application_actioned', 'application_requested', 'emoji_discovery', 'guild', 'keywords',
        'primary_category', 'sub_categories')
    def __init__(self, data, guild):
        """
        Creates a new ``GuildDiscovery`` from the requested data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Requested guild discovery data.
        guild : ``Guild``
            The owner guild.
        """
        self.guild = guild
        
        self._update_attributes(data)
    
    def _update_attributes(self, data):
        """
        Updates the guild discovery object from the given data by overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Guild discovery data.
        """
        self.sub_categories = set(DiscoveryCategory.from_id(category_id) for category_id in data['category_ids'])
        self.emoji_discovery = data['emoji_discoverability_enabled']
        
        keywords = data['keywords']
        if (keywords is not None):
            if keywords :
                keywords = set(keywords)
            else:
                keywords = None
        self.keywords = keywords
        
        self.primary_category = DiscoveryCategory.from_id(data['primary_category_id'], primary=True)
        
        application_actioned = data.get('partner_actioned_timestamp', None)
        if (application_actioned is not None):
            application_actioned = timestamp_to_datetime(application_actioned)
        
        self.application_actioned = application_actioned
        
        application_requested = data.get('partner_application_timestamp', None)
        if (application_requested is not None):
            application_requested = timestamp_to_datetime(application_requested)
        
        self.application_requested = application_requested
        
    def __eq__(self, other):
        """Returns whether the two guild discoveries are the same."""
        if (type(self) is not type(other)):
            return NotImplemented
        
        if (self.guild is not other.guild):
            return False
        
        if (self.primary_category != other.primary_category_id):
            return False
        
        if (self.emoji_discovery != other.emoji_discovery):
            return False
        
        # Leave the set-s last.
        if (self.sub_categories != other.category_ids):
            return False
        
        # Keywords can be `None`.
        if (self.keywords != other.keywords):
            return False
        
        if (self.application_actioned != other.application_actioned):
            return False
        
        return True
    
    def __hash__(self):
        """Returns the guild discovery's hash value."""
        return self.guild.id
    
    def __repr__(self):
        """Returns the guild discovery's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' of guild ',
        ]
        
        guild = self.guild
        repr_parts.append(repr(guild.name))
        repr_parts.append(' (')
        repr_parts.append(repr(guild.id))
        repr_parts.append(')>')
        
        return ''.join(repr_parts)


class DiscoveryCategory(DiscordEntity, immortal=True):
    """
    Represents a category of a ``GuildDiscovery``.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the discovery category.
    local_names : `None` or `dict` of (`str`, `str`) items
        The category's name in other languages.
    name : `str`
        The category's name.
    primary : `bool`
        Whether this category can be set as a guild's primary category.
    
    Class Attributes
    ----------------
    There are predefined discovery categories, which can be accessed as class attributes as well:
    
    +-------------------------------+-------+-------------------------------+-----------+
    | Class attribute name          | id    | name                          | primary   |
    +===============================+=======+===============================+===========+
    | general                       | `0`   | `'General'`                   | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | gaming                        | `1`   | `'Gaming'`                    | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | music                         | `2`   | `'Music'`                     | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | entertainment                 | `3`   | `'Entertainment'`             | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | creative_arts                 | `4`   | `'Creative Arts'`             | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | science_and_tech              | `5`   | `'Science & Tech'`            | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | education                     | `6`   | `'Education'`                 | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | sports                        | `7`   | `'Sports'`                    | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | fashion_and_beauty            | `8`   | `'Fashion & Beauty'`          | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | relationships_and_identity    | `9`   | `'Relationships & Identity'`  | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | travel_and_food               | `10`  | `'Travel & Food'`             | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | fitness_and_health            | `11`  | `'Fitness & Health'`          | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | finance                       | `12`  | `'Finance'`                   | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | other                         | `13`  | `'Other'`                     | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | general_chatting              | `14`  | `'General Chatting'`          | `True`    |
    +-------------------------------+-------+-------------------------------+-----------+
    | esports                       | `15`  | `'Esports'`                   | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | anime_and_manga               | `16`  | `'Anime & Manga'`             | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | movies_and_tv                 | `17`  | `'Movies & TV'`               | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | books                         | `18`  | `'Books'`                     | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | art                           | `19`  | `'Art'`                       | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | writing                       | `20`  | `'Writing'`                   | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | crafts_diy_and_making         | `21`  | `'Crafts, DIY, & Making'`     | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | programming                   | `22`  | `'Programming'`               | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | podcasts                      | `23`  | `'Podcasts'`                  | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | tabletop_games                | `24`  | `'Tabletop Games'`            | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | memes                         | `25`  | `'Memes'`                     | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | news_and_current_events       | `26`  | `'News & Current Events'`     | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | cryptocurrency                | `27`  | `'Cryptocurrency'`            | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | investing                     | `28`  | `'Investing'`                 | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | studying_and_teaching         | `29`  | `'Studying & Teaching'`       | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | lfg                           | `30`  | `'LFG'`                       | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | customer_support              | `31`  | `'Customer Support'`          | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | theorycraft                   | `32`  | `'Theorycraft'`               | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | events                        | `33`  | `'Events'`                    | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | roleplay                      | `34`  | `'Roleplay'`                  | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | content_creator               | `35`  | `'Content Creator'`           | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | business                      | `36`  | `'Business'`                  | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | local_group                   | `37`  | `'Local Group'`               | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | collaboration                 | `38`  | `'Collaboration'`             | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | fandom                        | `39`  | `'Fandom'`                    | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | wiki_and_guide                | `40`  | `'Wiki & Guide'`              | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | subreddit                     | `42`  | `'Subreddit'`                 | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | emoji                         | `43`  | `'Emoji'`                     | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | comics_and_cartoons           | `44`  | `'Comics & Cartoons'`         | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | mobile                        | `45`  | `'Mobile'`                    | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | console                       | `46`  | `'Console'`                   | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    | charity_and_nonprofit         | `47`  | `'Charity & Nonprofit'`       | `False`   |
    +-------------------------------+-------+-------------------------------+-----------+
    
    Notes
    -----
    Guild discovery objects are weakreferable.
    """
    __slots__ = ('local_names', 'name', 'primary')
    
    @classmethod
    def from_id(cls, category_id, primary=False):
        """
        Tries to find the category by the given id. If exists returns that, else creates a new one.
        
        Parameters
        ----------
        category_id : `int`
            The unique identifier number of the discovery category
        primary : `bool`, Optional
            Whether the category is a primary category.
        
        Returns
        -------
        category : ``DiscoveryCategory``
        """
        try:
            category = DISCOVERY_CATEGORIES[category_id]
        except KeyError:
            category = object.__new__(cls)
            DISCOVERY_CATEGORIES[category_id] = category
            category.id = category_id
            category.name = ''
            category.local_names = None
            category.primary = primary
        
        if primary:
            category.primary = True
        
        return category
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``DiscoveryCategory`` object from the given data. if the discovery category already
        exists returns that instead.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Discovery category data.
        
        Returns
        -------
        category : ``DiscoveryCategory``
        """
        # Note that even tho id is received as integer, but integers over 32 bits are still received as string.
        category_id = int(data['id'])
        
        try:
            category = DISCOVERY_CATEGORIES[category_id]
        except KeyError:
            category = object.__new__(cls)
            DISCOVERY_CATEGORIES[category_id] = category
            category.id = category_id
        
        category.primary = data['is_primary']
        name_data = data['name']
        category.name = name_data['default']
        
        local_names = name_data.get('localizations', None)
        if (local_names is not None) and (not local_names):
            local_names = None
        
        category.local_names = local_names
        
        return category
    
    def __init__(self, category_id, name, primary):
        """
        Creates a new discovery category from the given parameters.
        
        Parameters
        ----------
        category_id : `int`
            The unique identifier number of the discovery category.
        name : `str`
            The category's name.
        primary : `bool`
            Whether this category can be set as a guild's primary category.
        """
        self.id = category_id
        self.name = name
        self.primary = primary
        self.local_names = None
        
        DISCOVERY_CATEGORIES[category_id] = self
    
    def created_at(self):
        """
        Returns when the discovery category was created.
        
        Because discovery category id-s are not snowflakes, this method will return the start of the discord epoch.
        
        Returns
        -------
        created_at : `datetime`
        """
        return DISCORD_EPOCH_START
    
    
    def __repr__(self):
        """Returns the discovery category's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' name=',
            repr(self.name),
            ' id=',
            repr(self.id),
        ]
        
        if self.primary:
            repr_parts.append(' primary')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    general = NotImplemented
    gaming = NotImplemented
    music = NotImplemented
    entertainment = NotImplemented
    creative_arts = NotImplemented
    science_and_tech = NotImplemented
    education = NotImplemented
    sports = NotImplemented
    fashion_and_beauty = NotImplemented
    relationships_and_identity = NotImplemented
    travel_and_food = NotImplemented
    fitness_and_health = NotImplemented
    finance = NotImplemented
    other = NotImplemented
    general_chatting = NotImplemented
    esports = NotImplemented
    anime_and_manga = NotImplemented
    movies_and_tv = NotImplemented
    books = NotImplemented
    art = NotImplemented
    writing = NotImplemented
    crafts_diy_and_making = NotImplemented
    programming = NotImplemented
    podcasts = NotImplemented
    tabletop_games = NotImplemented
    memes = NotImplemented
    news_and_current_events = NotImplemented
    cryptocurrency = NotImplemented
    investing = NotImplemented
    studying_and_teaching = NotImplemented
    lfg = NotImplemented
    customer_support = NotImplemented
    theorycraft = NotImplemented
    events = NotImplemented
    roleplay = NotImplemented
    content_creator = NotImplemented
    business = NotImplemented
    local_group = NotImplemented
    collaboration = NotImplemented
    fandom = NotImplemented
    wiki_and_guide = NotImplemented
    subreddit = NotImplemented
    emoji = NotImplemented
    comics_and_cartoons = NotImplemented
    mobile = NotImplemented
    console = NotImplemented
    charity_and_nonprofit = NotImplemented

DiscoveryCategory.general = DiscoveryCategory(0, 'General', True)
DiscoveryCategory.gaming = DiscoveryCategory(1, 'Gaming', True)
DiscoveryCategory.music = DiscoveryCategory(2, 'Music', True)
DiscoveryCategory.entertainment = DiscoveryCategory(3, 'Entertainment', True)
DiscoveryCategory.creative_arts = DiscoveryCategory(4, 'Creative Arts', True)
DiscoveryCategory.science_and_tech = DiscoveryCategory(5, 'Science & Tech', True)
DiscoveryCategory.education = DiscoveryCategory(6, 'Education', True)
DiscoveryCategory.sports = DiscoveryCategory(7, 'Sports', True)
DiscoveryCategory.fashion_and_beauty = DiscoveryCategory(8, 'Fashion & Beauty', True)
DiscoveryCategory.relationships_and_identity = DiscoveryCategory(9, 'Relationships & Identity', True)
DiscoveryCategory.travel_and_food = DiscoveryCategory(10, 'Travel & Food', True)
DiscoveryCategory.fitness_and_health = DiscoveryCategory(11, 'Fitness & Health', True)
DiscoveryCategory.finance = DiscoveryCategory(12, 'Finance', True)
DiscoveryCategory.other = DiscoveryCategory(13, 'Other', True)
DiscoveryCategory.general_chatting = DiscoveryCategory(14, 'General Chatting', True)
DiscoveryCategory.esports = DiscoveryCategory(15, 'Esports', False)
DiscoveryCategory.anime_and_manga = DiscoveryCategory(16, 'Anime & Manga', False)
DiscoveryCategory.movies_and_tv = DiscoveryCategory(17, 'Movies & TV', False)
DiscoveryCategory.books = DiscoveryCategory(18, 'Books', False)
DiscoveryCategory.art = DiscoveryCategory(19, 'Art', False)
DiscoveryCategory.writing = DiscoveryCategory(20, 'Writing', False)
DiscoveryCategory.crafts_diy_and_making = DiscoveryCategory(21, 'Crafts, DIY, & Making', False)
DiscoveryCategory.programming = DiscoveryCategory(22, 'Programming', False)
DiscoveryCategory.podcasts = DiscoveryCategory(23, 'Podcasts', False)
DiscoveryCategory.tabletop_games = DiscoveryCategory(24, 'Tabletop Games', False)
DiscoveryCategory.memes = DiscoveryCategory(25, 'Memes', False)
DiscoveryCategory.news_and_current_events = DiscoveryCategory(26, 'News & Current Events', False)
DiscoveryCategory.cryptocurrency = DiscoveryCategory(27, 'Cryptocurrency', False)
DiscoveryCategory.investing = DiscoveryCategory(28, 'Investing', False)
DiscoveryCategory.studying_and_teaching = DiscoveryCategory(29, 'Studying & Teaching', False)
DiscoveryCategory.lfg = DiscoveryCategory(30, 'LFG', False)
DiscoveryCategory.customer_support = DiscoveryCategory(31, 'Customer Support', False)
DiscoveryCategory.theorycraft = DiscoveryCategory(32, 'Theorycraft', False)
DiscoveryCategory.events = DiscoveryCategory(33, 'Events', False)
DiscoveryCategory.roleplay = DiscoveryCategory(34, 'Roleplay', False)
DiscoveryCategory.content_creator = DiscoveryCategory(35, 'Content Creator', False)
DiscoveryCategory.business = DiscoveryCategory(36, 'Business', False)
DiscoveryCategory.local_group = DiscoveryCategory(37, 'Local Group', False)
DiscoveryCategory.collaboration = DiscoveryCategory(38, 'Collaboration', False)
DiscoveryCategory.fandom = DiscoveryCategory(39, 'Fandom', False)
DiscoveryCategory.wiki_and_guide = DiscoveryCategory(40, 'Wiki & Guide', False)
DiscoveryCategory.subreddit = DiscoveryCategory(42, 'Subreddit', False)
DiscoveryCategory.emoji = DiscoveryCategory(43, 'Emoji', False)
DiscoveryCategory.comics_and_cartoons = DiscoveryCategory(44, 'Comics & Cartoons', False)
DiscoveryCategory.mobile = DiscoveryCategory(45, 'Mobile', False)
DiscoveryCategory.console = DiscoveryCategory(46, 'Console', False)
DiscoveryCategory.charity_and_nonprofit = DiscoveryCategory(47, 'Charity & Nonprofit', False)
